###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# ruff: noqa: N802
"""Sqla query builder implementation"""

import uuid
import warnings
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from sqlalchemy import and_, not_, or_
from sqlalchemy import func as sa_func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.engine import Row
from sqlalchemy.exc import CompileError, SAWarning
from sqlalchemy.orm import aliased
from sqlalchemy.orm.attributes import InstrumentedAttribute, QueryableAttribute
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.elements import BinaryExpression, BooleanClauseList, Cast, ColumnClause, ColumnElement, Label
from sqlalchemy.sql.expression import case, text
from sqlalchemy.types import Boolean, DateTime, Float, Integer, String

from aiida.common.exceptions import NotExistent
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation.querybuilder import QUERYBUILD_LOGGER, BackendQueryBuilder, QueryDictType

from .joiner import JoinReturn, SqlaJoiner

jsonb_typeof = sa_func.jsonb_typeof
jsonb_array_length = sa_func.jsonb_array_length
array_length = sa_func.array_length

PROJECT_MAP = {
    'db_dbauthinfo': {
        'pk': 'id',
        'computer_pk': 'dbcomputer_id',
        'user_pk': 'aiidauser_id',
    },
    'db_dbnode': {
        'pk': 'id',
        'dict': 'attributes',
        'computer_pk': 'dbcomputer_id',
        'user_pk': 'user_id',
    },
    'db_dbcomputer': {
        'pk': 'id',
    },
    'db_dbgroup': {
        'pk': 'id',
        'user_pk': 'user_id',
    },
    'db_dbcomment': {
        'pk': 'id',
        'user_pk': 'user_id',
        'node_pk': 'dbnode_id',
    },
    'db_dblog': {
        'pk': 'id',
        'node_pk': 'dbnode_id',
    },
}


@dataclass
class BuiltQuery:
    """A class to store the query and the corresponding projections."""

    query: Query
    tag_to_alias: Dict[str, Optional[AliasedClass]]
    tag_to_projected: Dict[str, Dict[str, int]]


class SqlaQueryBuilder(BackendQueryBuilder):
    """QueryBuilder to use with SQLAlchemy-backend and
    schema defined in backends.sqlalchemy.models
    """

    def __init__(self, backend):
        super().__init__(backend)

        self._joiner = SqlaJoiner(self, self.build_filters)

        # Set conversions between the field names in the database and used by the `QueryBuilder`
        # table -> field -> field
        self.inner_to_outer_schema: Dict[str, Dict[str, str]] = {
            'db_dbauthinfo': {'_metadata': 'metadata'},
            'db_dbcomputer': {'_metadata': 'metadata'},
            'db_dblog': {'_metadata': 'metadata'},
        }
        self.outer_to_inner_schema: Dict[str, Dict[str, str]] = {
            'db_dbauthinfo': {'metadata': '_metadata'},
            'db_dbcomputer': {'metadata': '_metadata'},
            'db_dblog': {'metadata': '_metadata'},
        }

        # data generated from front-end
        self._data: QueryDictType = {
            'path': [],
            'filters': {},
            'project': {},
            'project_map': {},
            'order_by': [],
            'offset': None,
            'limit': None,
            'distinct': False,
        }

        # Hashing the internal query representation avoids rebuilding a query
        self._query_cache: Optional[BuiltQuery] = None
        self._query_hash: Optional[str] = None

    @property
    def Node(self):
        import aiida.storage.psql_dos.models.node

        return aiida.storage.psql_dos.models.node.DbNode

    @property
    def Link(self):
        import aiida.storage.psql_dos.models.node

        return aiida.storage.psql_dos.models.node.DbLink

    @property
    def Computer(self):
        import aiida.storage.psql_dos.models.computer

        return aiida.storage.psql_dos.models.computer.DbComputer

    @property
    def User(self):
        import aiida.storage.psql_dos.models.user

        return aiida.storage.psql_dos.models.user.DbUser

    @property
    def Group(self):
        import aiida.storage.psql_dos.models.group

        return aiida.storage.psql_dos.models.group.DbGroup

    @property
    def AuthInfo(self):
        import aiida.storage.psql_dos.models.authinfo

        return aiida.storage.psql_dos.models.authinfo.DbAuthInfo

    @property
    def Comment(self):
        import aiida.storage.psql_dos.models.comment

        return aiida.storage.psql_dos.models.comment.DbComment

    @property
    def Log(self):
        import aiida.storage.psql_dos.models.log

        return aiida.storage.psql_dos.models.log.DbLog

    @property
    def table_groups_nodes(self):
        import aiida.storage.psql_dos.models.group

        return aiida.storage.psql_dos.models.group.table_groups_nodes

    def get_session(self) -> Session:
        """Get the connection to the database"""
        return self._backend.get_session()

    def count(self, data: QueryDictType) -> int:
        with self.query_session(data) as build:
            result = build.query.count()
        return result

    def first(self, data: QueryDictType) -> Optional[List[Any]]:
        with self.query_session(data) as build:
            result = build.query.first()

        if result is None:
            return result

        # SQLA will return a Row, if only certain columns are requested,
        # or a database model if that is requested
        if not isinstance(result, Row):
            result = (result,)

        return [self.to_backend(r) for r in result]

    def iterall(self, data: QueryDictType, batch_size: Optional[int]) -> Iterable[List[Any]]:
        """Return an iterator over all the results of a list of lists."""
        with self.query_session(data) as build:
            stmt = build.query.statement.execution_options(yield_per=batch_size)
            session = self.get_session()

            # Open a session transaction unless already inside one. This prevents the `ModelWrapper` from calling commit
            # on the session when a yielded row is mutated. This would reset the cursor invalidating it and causing an
            # exception to be raised in the next batch of rows in the iteration.
            # See https://github.com/python/mypy/issues/10109 for the reason of the type warning.
            with nullcontext() if session.in_nested_transaction() else self._backend.transaction():  # type: ignore[attr-defined]
                for resultrow in session.execute(stmt):
                    yield [self.to_backend(rowitem) for rowitem in resultrow]

    def iterdict(self, data: QueryDictType, batch_size: Optional[int]) -> Iterable[Dict[str, Dict[str, Any]]]:
        """Return an iterator over all the results of a list of dictionaries."""
        with self.query_session(data) as build:
            stmt = build.query.statement.execution_options(yield_per=batch_size)
            session = self.get_session()

            # Open a session transaction unless already inside one. This prevents the `ModelWrapper` from calling commit
            # on the session when a yielded row is mutated. This would reset the cursor invalidating it and causing an
            # exception to be raised in the next batch of rows in the iteration.
            # See https://github.com/python/mypy/issues/10109 for the reason of the type warning.
            with nullcontext() if session.in_nested_transaction() else self._backend.transaction():  # type: ignore[attr-defined]
                for row in self.get_session().execute(stmt):
                    # build the yield result
                    yield_result: Dict[str, Dict[str, Any]] = {}
                    for (
                        tag,
                        projected_entities_dict,
                    ) in build.tag_to_projected.items():
                        yield_result[tag] = {}
                        for attrkey, project_index in projected_entities_dict.items():
                            alias = build.tag_to_alias.get(tag)
                            if alias is None:
                                raise ValueError(f'No alias found for tag {tag}')
                            field_name = get_corresponding_property(
                                get_table_name(alias),
                                attrkey,
                                self.inner_to_outer_schema,
                            )
                            key = self._data['project_map'].get(tag, {}).get(field_name, field_name)
                            yield_result[tag][key] = self.to_backend(row[project_index])
                    yield yield_result

    def get_query(self, data: QueryDictType) -> BuiltQuery:
        """Return the built query.

        To avoid unnecessary re-builds of the query, the hashed dictionary representation of this instance
        is compared to the last query returned, which is cached by its hash.
        """
        from aiida.common.hashing import make_hash

        query_hash = make_hash(data)

        if not (self._query_cache and self._query_hash and self._query_hash == query_hash):
            self._query_cache = self._build(data)
            self._query_hash = query_hash

        return self._query_cache

    @contextmanager
    def query_session(self, data: QueryDictType) -> Iterator[BuiltQuery]:
        """Yield the built query, ensuring the session is closed on an exception."""
        query = self.get_query(data)
        try:
            yield query
        except Exception:
            self.get_session().close()
            raise

    def _build(self, data: QueryDictType) -> BuiltQuery:
        """Build the query and return."""
        # generate aliases for tags
        tag_to_alias: Dict[str, Optional[AliasedClass]] = {}
        cls_map = {
            EntityTypes.AUTHINFO.value: self.AuthInfo,
            EntityTypes.COMMENT.value: self.Comment,
            EntityTypes.COMPUTER.value: self.Computer,
            EntityTypes.GROUP.value: self.Group,
            EntityTypes.NODE.value: self.Node,
            EntityTypes.LOG.value: self.Log,
            EntityTypes.USER.value: self.User,
            EntityTypes.LINK.value: self.Link,
        }
        for path in data['path']:
            # An SAWarning warning is currently emitted:
            # "relationship 'DbNode.input_links' will copy column db_dbnode.id to column db_dblink.output_id,
            # which conflicts with relationship(s): 'DbNode.outputs' (copies db_dbnode.id to db_dblink.output_id)"
            # This should be eventually fixed
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=SAWarning)
                tag_to_alias[path['tag']] = aliased(cls_map[path['orm_base']])

        # now create joins first, so we can populate edge tag aliases
        joins = generate_joins(data, tag_to_alias, self._joiner)
        for join in joins:
            if join.aliased_edge is not None:
                tag_to_alias[join.edge_tag] = join.aliased_edge

        # generate the projections
        projections, tag_to_projected = generate_projections(
            data, tag_to_alias, self.outer_to_inner_schema, self._get_projectable_entity
        )

        # initialise the query
        query = self.get_session().query()
        # add the projections
        for projection, is_entity in projections:
            if is_entity:
                query = query.add_entity(projection)
            else:
                query = query.add_columns(projection)
        # and the starting table, from which joins will be made
        starting_table = tag_to_alias.get(data['path'][0]['tag'])
        if starting_table is None:
            raise ValueError('starting tag not found')
        query = query.select_from(starting_table)

        # add the joins
        for join in joins:
            query = join.join(query)

        # add the filters
        for tag, filter_specs in data['filters'].items():
            if not filter_specs:
                continue
            alias = tag_to_alias.get(tag)
            if not alias:
                raise ValueError(f'Unknown tag {tag!r} in filters, known: {list(tag_to_alias)}')
            filters = self.build_filters(alias, filter_specs)
            if filters is not None:
                query = query.filter(filters)

        # set the ordering
        for order_spec in data['order_by']:
            for tag, entity_list in order_spec.items():
                alias = tag_to_alias.get(tag)
                if not alias:
                    raise ValueError(f'Unknown tag {tag!r} in order_by, known: {list(tag_to_alias)}')
                for entitydict in entity_list:
                    for entitytag, entityspec in entitydict.items():
                        query = query.order_by(self._create_order_by(alias, entitytag, entityspec))

        # final setup for the query
        if data['limit'] is not None:
            query = query.limit(data['limit'])
        if data['offset'] is not None:
            query = query.offset(data['offset'])
        if data['distinct']:
            query = query.distinct()

        return BuiltQuery(query, tag_to_alias, tag_to_projected)

    def _create_order_by(
        self, alias: AliasedClass, field_key: str, entityspec: dict
    ) -> Union[ColumnElement, InstrumentedAttribute]:
        """Build the order_by parameter of the query."""
        column_name = field_key.split('.')[0]
        attrpath = field_key.split('.')[1:]
        if attrpath and 'cast' not in entityspec.keys():
            # JSONB fields ar delimited by '.' must be cast
            raise ValueError(
                f'To order_by {field_key!r}, the value has to be cast, ' "but no 'cast' key has been specified."
            )
        entity = self._get_projectable_entity(alias, column_name, attrpath, cast=entityspec.get('cast'))
        order = entityspec.get('order', 'asc')
        if order == 'desc':
            entity = entity.desc()
        elif order != 'asc':
            raise ValueError(f"Unknown 'order' key: {order!r}, must be one of: 'asc', 'desc'")
        return entity

    @staticmethod
    def _get_projectable_entity(
        alias: AliasedClass,
        column_name: str,
        attrpath: List[str],
        cast: Optional[str] = None,
    ) -> Union[ColumnElement, InstrumentedAttribute]:
        """Return projectable entity for a given alias and column name."""
        if not (attrpath or column_name in ('attributes', 'extras')):
            return get_column(column_name, alias)

        entity: ColumnElement = get_column(column_name, alias)[attrpath]
        if cast is None:
            pass
        elif cast == 'f':
            entity = entity.astext.cast(Float)
        elif cast == 'i':
            entity = entity.astext.cast(Integer)
        elif cast == 'b':
            entity = entity.astext.cast(Boolean)
        elif cast == 't':
            entity = entity.astext
        elif cast == 'j':
            entity = entity.astext.cast(JSONB)
        elif cast == 'd':
            entity = entity.astext.cast(DateTime)
        else:
            raise ValueError(f'Unknown casting key {cast}')
        return entity

    def build_filters(self, alias: AliasedClass, filter_spec: Dict[str, Any]) -> Optional[BooleanClauseList]:
        """Recurse through the filter specification and apply filter operations.

        :param alias: The alias of the ORM class the filter will be applied on
        :param filter_spec: the specification of the filter

        :returns: an sqlalchemy expression.
        """
        expressions: List[Any] = []
        for path_spec, filter_operation_dict in filter_spec.items():
            if path_spec in ('and', 'or', '~or', '~and', '!and', '!or'):
                subexpressions = []
                for sub_filter_spec in filter_operation_dict:
                    filters = self.build_filters(alias, sub_filter_spec)
                    if filters is not None:
                        subexpressions.append(filters)
                if subexpressions:
                    if path_spec == 'and':
                        expressions.append(and_(*subexpressions))
                    elif path_spec == 'or':
                        expressions.append(or_(*subexpressions))
                    elif path_spec in ('~and', '!and'):
                        expressions.append(not_(and_(*subexpressions)))
                    elif path_spec in ('~or', '!or'):
                        expressions.append(not_(or_(*subexpressions)))
            else:
                column_name = path_spec.split('.')[0]

                attr_key = path_spec.split('.')[1:]
                is_jsonb = bool(attr_key) or column_name in ('dict', 'attributes', 'extras')
                column: Optional[InstrumentedAttribute]
                try:
                    column = get_column(column_name, alias)
                except (ValueError, TypeError):
                    if is_jsonb:
                        column = None
                    else:
                        raise
                if not isinstance(filter_operation_dict, dict):
                    filter_operation_dict = {'==': filter_operation_dict}  # noqa: PLW2901
                for operator, value in filter_operation_dict.items():
                    expressions.append(
                        self.get_filter_expr(
                            operator,
                            value,
                            attr_key,
                            is_jsonb=is_jsonb,
                            column=column,
                            column_name=column_name,
                            alias=alias,
                        )
                    )
        return and_(*expressions) if expressions else None

    def get_filter_expr(
        self,
        operator: str,
        value: Any,
        attr_key: List[str],
        is_jsonb: bool,
        alias=None,
        column=None,
        column_name=None,
    ):
        """Applies a filter on the alias given.

        Expects the alias of the ORM-class on which to filter, and filter_spec.
        Filter_spec contains the specification on the filter.
        Expects:

        :param operator: The operator to apply, see below for further details
        :param value:
            The value for the right side of the expression,
            the value you want to compare with.

        :param path: The path leading to the value

        :param is_jsonb: Whether the value is in a json-column, or in an attribute like table.


        Implemented and valid operators:

        *   for any type:
            *   ==  (compare single value, eg: '==':5.0)
            *   in    (compare whether in list, eg: 'in':[5, 6, 34]
        *  for floats and integers:
            *   >
            *   <
            *   <=
            *   >=
        *  for strings:
            *   like  (case - sensitive), for example
                'like':'node.calc.%'  will match node.calc.relax and
                node.calc.RELAX and node.calc. but
                not node.CALC.relax
            *   ilike (case - unsensitive)
                will also match node.CaLc.relax in the above example

            .. note::
                The character % is a reserved special character in SQL,
                and acts as a wildcard. If you specifically
                want to capture a ``%`` in the string, use: ``_%``

        *   for arrays and dictionaries (only for the
            SQLAlchemy implementation):

            *   contains: pass a list with all the items that
                the array should contain, or that should be among
                the keys, eg: 'contains': ['N', 'H'])
            *   has_key: pass an element that the list has to contain
                or that has to be a key, eg: 'has_key':'N')

        *  for arrays only (SQLAlchemy version):
            *   of_length
            *   longer
            *   shorter

        All the above filters invoke a negation of the
        expression if preceded by **~**::

            # first example:
            filter_spec = {
                'name' : {
                    '~in':[
                        'halle',
                        'lujah'
                    ]
                } # Name not 'halle' or 'lujah'
            }

            # second example:
            filter_spec =  {
                'id' : {
                    '~==': 2
                }
            } # id is not 2
        """
        expr: Any = None
        if operator.startswith('~'):
            negation = True
            operator = operator.lstrip('~')
        elif operator.startswith('!'):
            negation = True
            operator = operator.lstrip('!')
        else:
            negation = False
        if operator in ('longer', 'shorter', 'of_length'):
            if not isinstance(value, int):
                raise TypeError('You have to give an integer when comparing to a length')
        elif operator in ('like', 'ilike'):
            if not isinstance(value, str):
                raise TypeError(f'Value for operator {operator} has to be a string (you gave {value})')

        elif operator == 'in':
            try:
                value_type_set = set(type(i) for i in value)
            except TypeError:
                raise TypeError('Value for operator `in` could not be iterated')
            if not value_type_set:
                raise ValueError('Value for operator `in` is an empty list')
            if len(value_type_set) > 1:
                raise ValueError(f'Value for operator `in` contains more than one type: {value}')
        elif operator in ('and', 'or'):
            expressions_for_this_path = []
            for filter_operation_dict in value:
                for newoperator, newvalue in filter_operation_dict.items():
                    expressions_for_this_path.append(
                        self.get_filter_expr(
                            newoperator,
                            newvalue,
                            attr_key=attr_key,
                            is_jsonb=is_jsonb,
                            alias=alias,
                            column=column,
                            column_name=column_name,
                        )
                    )
            if operator == 'and':
                expr = and_(*expressions_for_this_path)
            elif operator == 'or':
                expr = or_(*expressions_for_this_path)

        if expr is None:
            if is_jsonb:
                expr = self.get_filter_expr_from_jsonb(
                    operator,
                    value,
                    attr_key,
                    column=column,
                    column_name=column_name,
                    alias=alias,
                )
            else:
                if column is None:
                    if (alias is None) and (column_name is None):
                        raise RuntimeError('I need to get the column but do not know the alias and the column name')
                    column = get_column(column_name, alias)
                expr = self.get_filter_expr_from_column(operator, value, column)

        if negation:
            return not_(expr)
        return expr

    @staticmethod
    def get_filter_expr_from_jsonb(
        operator: str,
        value,
        attr_key: List[str],
        column=None,
        column_name=None,
        alias=None,
    ):
        """Return a filter expression"""

        def cast_according_to_type(path_in_json, value):
            """Cast the value according to the type"""
            if isinstance(value, bool):
                type_filter = jsonb_typeof(path_in_json) == 'boolean'
                casted_entity = path_in_json.astext.cast(Boolean)
            elif isinstance(value, (int, float)):
                type_filter = jsonb_typeof(path_in_json) == 'number'
                casted_entity = path_in_json.astext.cast(Float)
            elif isinstance(value, dict) or value is None:
                type_filter = jsonb_typeof(path_in_json) == 'object'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, dict):
                type_filter = jsonb_typeof(path_in_json) == 'array'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, str):
                type_filter = jsonb_typeof(path_in_json) == 'string'
                casted_entity = path_in_json.astext
            elif value is None:
                type_filter = jsonb_typeof(path_in_json) == 'null'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            else:
                raise TypeError(f'Unknown type {type(value)}')
            return type_filter, casted_entity

        if column is None:
            column = get_column(column_name, alias)

        database_entity = column[tuple(attr_key)]
        expr: Any
        if operator == '==':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity == value), else_=False)
        elif operator == '>':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity > value), else_=False)
        elif operator == '<':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity < value), else_=False)
        elif operator in ('>=', '=>'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity >= value), else_=False)
        elif operator in ('<=', '=<'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity <= value), else_=False)
        elif operator == 'of_type':
            # http://www.postgresql.org/docs/9.5/static/functions-json.html
            #  Possible types are object, array, string, number, boolean, and null.
            valid_types = ('object', 'array', 'string', 'number', 'boolean', 'null')
            if value not in valid_types:
                raise ValueError(f'value {value} for of_type is not among valid types\n{valid_types}')
            expr = jsonb_typeof(database_entity) == value
        elif operator == 'like':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity.like(value)), else_=False)
        elif operator == 'ilike':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case((type_filter, casted_entity.ilike(value)), else_=False)
        elif operator == 'in':
            type_filter, casted_entity = cast_according_to_type(database_entity, value[0])
            expr = case((type_filter, casted_entity.in_(value)), else_=False)
        elif operator == 'contains':
            expr = database_entity.cast(JSONB).contains(value)
        elif operator == 'has_key':
            expr = database_entity.cast(JSONB).has_key(value)
        elif operator == 'of_length':
            expr = case(
                (
                    jsonb_typeof(database_entity) == 'array',
                    jsonb_array_length(database_entity.cast(JSONB)) == value,
                ),
                else_=False,
            )

        elif operator == 'longer':
            expr = case(
                (
                    jsonb_typeof(database_entity) == 'array',
                    jsonb_array_length(database_entity.cast(JSONB)) > value,
                ),
                else_=False,
            )
        elif operator == 'shorter':
            expr = case(
                (
                    jsonb_typeof(database_entity) == 'array',
                    jsonb_array_length(database_entity.cast(JSONB)) < value,
                ),
                else_=False,
            )
        else:
            raise ValueError(f'Unknown operator {operator} for filters in JSON field')
        return expr

    @staticmethod
    def get_filter_expr_from_column(operator: str, value: Any, column) -> BinaryExpression:
        """A method that returns an valid SQLAlchemy expression.

        :param operator: The operator provided by the user ('==',  '>', ...)
        :param value: The value to compare with, e.g. (5.0, 'foo', ['a','b'])
        :param column: an instance of sqlalchemy.orm.attributes.InstrumentedAttribute or

        """
        # Label is used because it is what is returned for the
        # 'state' column by the hybrid_column construct
        if not isinstance(
            column,
            (Cast, InstrumentedAttribute, QueryableAttribute, Label, ColumnClause),
        ):
            raise TypeError(f'column ({type(column)}) {column} is not a valid column')
        database_entity = column
        if operator == '==':
            expr = database_entity == value
        elif operator == '>':
            expr = database_entity > value
        elif operator == '<':
            expr = database_entity < value
        elif operator == '>=':
            expr = database_entity >= value
        elif operator == '<=':
            expr = database_entity <= value
        elif operator == 'like':
            # the like operator expects a string, so we cast to avoid problems
            # with fields like UUID, which don't support the like operator
            expr = database_entity.cast(String).like(value)
        elif operator == 'ilike':
            expr = database_entity.ilike(value)
        elif operator == 'in':
            expr = database_entity.in_(value)
        else:
            raise ValueError(f'Unknown operator {operator} for filters on columns')
        return expr

    def to_backend(self, res) -> Any:
        """Convert results to return backend specific objects.

        - convert `DbModel` instances to `BackendEntity` instances.
        - convert UUIDs to strings

        :param res: the result returned by the query

        :returns:backend compatible instance
        """
        if isinstance(res, uuid.UUID):
            return str(res)

        try:
            return self._backend.get_backend_entity(res)
        except TypeError:
            return res

    def as_sql(self, data: QueryDictType, inline: bool = False) -> str:
        """Return the SQL query as a string."""
        with self.query_session(data) as build:
            compiled = compile_query(build.query, literal_binds=inline)
        if inline:
            return compiled.string + '\n'
        return f'{compiled.string!r} % {compiled.params!r}\n'

    def analyze_query(self, data: QueryDictType, execute: bool = True, verbose: bool = False) -> str:
        """Analyze the query and return the result as a string."""
        with self.query_session(data) as build:
            if build.query.session.bind.dialect.name != 'postgresql':  # type: ignore[union-attr]
                raise NotImplementedError('Only PostgreSQL is supported for this method')
            compiled = compile_query(build.query, literal_binds=True)
        options = ', '.join((['ANALYZE'] if execute else []) + (['VERBOSE'] if verbose else []))
        options = f' ({options})' if options else ''
        rows = self.get_session().execute(text(f'EXPLAIN{options} {compiled.string}')).fetchall()
        return '\n'.join(row[0] for row in rows)

    def get_creation_statistics(self, user_pk: Optional[int] = None) -> Dict[str, Any]:
        session = self.get_session()
        retdict: Dict[Any, Any] = {}

        total_query = session.query(self.Node)
        types_query = session.query(self.Node.node_type.label('typestring'), sa_func.count(self.Node.id))
        stat_query = session.query(
            sa_func.date_trunc('day', self.Node.ctime).label('cday'),
            sa_func.count(self.Node.id),
        )

        if user_pk is not None:
            total_query = total_query.filter(self.Node.user_id == user_pk)
            types_query = types_query.filter(self.Node.user_id == user_pk)
            stat_query = stat_query.filter(self.Node.user_id == user_pk)

        # Total number of nodes
        retdict['total'] = total_query.count()

        # Nodes per type
        retdict['types'] = dict(types_query.group_by('typestring').all())

        # Nodes created per day
        stat = stat_query.group_by('cday').order_by('cday').all()

        ctime_by_day = {_[0].strftime('%Y-%m-%d'): _[1] for _ in stat}
        retdict['ctime_by_day'] = ctime_by_day

        return retdict
        # Still not containing all dates


def get_corresponding_property(entity_table: str, given_property: str, mapper: Dict[str, Dict[str, str]]) -> str:
    """This method returns an updated property for a given a property.
    If there is no update for the property, the given property is returned.
    """
    try:
        # Get the mapping for the specific entity_table
        property_mapping = mapper[entity_table]
        try:
            # Get the mapping for the specific property
            return property_mapping[given_property]
        except KeyError:
            # If there is no mapping, the property remains unchanged
            return given_property
    except KeyError:
        # If it doesn't exist, it means that the given_property remains v
        return given_property


def get_corresponding_properties(entity_table: str, given_properties: List[str], mapper: Dict[str, Dict[str, str]]):
    """This method returns a list of updated properties for a given list of properties.
    If there is no update for the property, the given property is returned in the list.
    """
    if entity_table in mapper:
        res = []
        for given_property in given_properties:
            res.append(get_corresponding_property(entity_table, given_property, mapper))
        return res

    return given_properties


def modify_expansions(
    alias: AliasedClass,
    expansions: List[str],
    outer_to_inner_schema=Dict[str, Dict[str, str]],
) -> List[str]:
    """Modify names of projections if `**` was specified.

    This is important for the schema having attributes in a different table.
    In SQLA, the metadata should be changed to _metadata to be in-line with the database schema
    """
    # The following check is added to avoided unnecessary calls to get_inner_property for QB edge queries
    # The update of expansions makes sense only when AliasedClass is provided
    if hasattr(alias, '_sa_class_manager'):
        if '_metadata' in expansions:
            raise NotExistent(f"_metadata doesn't exist for {alias}. Please try metadata.")

        return get_corresponding_properties(alias.__tablename__, expansions, outer_to_inner_schema)

    return expansions


def get_column(colname: str, alias: AliasedClass) -> InstrumentedAttribute:
    """Return the column for a given projection."""
    table_name = get_table_name(alias)
    if projections := PROJECT_MAP.get(table_name):
        colname = projections.get(colname, colname)
    try:
        return getattr(alias, colname)
    except AttributeError as exc:
        raise ValueError(
            '{} is not a column of {}\n' 'Valid columns are:\n' '{}'.format(
                colname, alias, '\n'.join(alias._sa_class_manager.mapper.c.keys())
            )
        ) from exc


def get_column_names(alias: AliasedClass) -> List[str]:
    """Given the backend specific alias, return the column names that correspond to the aliased table."""
    return [str(c).replace(f'{alias.__table__.name}.', '') for c in alias.__table__.columns]


def get_table_name(aliased_class: AliasedClass) -> str:
    """Returns the table name given an Aliased class"""
    return aliased_class.__tablename__


def compile_query(query: Query, literal_binds: bool = False) -> SQLCompiler:
    """Compile the query to the SQL executable.

    :params literal_binds: Inline bound parameters (this is normally handled by the Python DBAPI).
    """
    dialect = query.session.bind.dialect  # type: ignore[union-attr]

    class _Compiler(dialect.statement_compiler):  # type: ignore[name-defined]
        """Override the compiler with additional literal value renderers."""

        def render_literal_value(self, value, type_):
            """Render the value of a bind parameter as a quoted literal.

            See https://www.postgresql.org/docs/current/functions-json.html for serialisation specs
            """
            from datetime import date, datetime, timedelta

            try:
                return super().render_literal_value(value, type_)
            # sqlalchemy<1.4.45 raises NotImplementedError, sqlalchemy>=1.4.45 raises CompileError
            except (NotImplementedError, CompileError):
                if isinstance(value, list):
                    values = ','.join(self.render_literal_value(item, type_) for item in value)
                    return f"'[{values}]'"
                if isinstance(value, int):
                    return str(value)
                if isinstance(value, (str, date, datetime, timedelta)):
                    escaped = str(value).replace('"', '\\"')
                    return f'"{escaped}"'
                raise

    return _Compiler(dialect, query.statement, compile_kwargs=dict(literal_binds=literal_binds))


def generate_joins(
    data: QueryDictType, aliases: Dict[str, Optional[AliasedClass]], joiner: SqlaJoiner
) -> List[JoinReturn]:
    """Generate the joins for the query."""
    joins: List[JoinReturn] = []
    # Start on second path item, since there is nothing to join if that is the first table
    for index, verticespec in enumerate(data['path'][1:], start=1):
        join_to = aliases[verticespec['tag']]
        if join_to is None:
            raise ValueError(f'No alias found for tag {verticespec["tag"]}')

        calling_entity = data['path'][index]['orm_base']
        joining_keyword = verticespec['joining_keyword']
        joining_value = verticespec['joining_value']
        try:
            join_func = joiner.get_join_func(calling_entity, joining_keyword)
        except KeyError:
            raise ValueError(f"'{joining_keyword}' is not a valid joining keyword for a '{calling_entity}' type entity")

        if not isinstance(joining_value, str):
            raise ValueError(f"'joining_value' value is not a string: {joining_value}")

        join_tag = aliases.get(joining_value, None)
        if not join_tag:
            raise ValueError(f'no alias found for joining_value tag {joining_value!r}')

        edge_tag = verticespec['edge_tag']

        # if verticespec['joining_keyword'] in ('with_ancestors', 'with_descendants'):
        # These require a filter_dict, to help the recursive function find a good starting point.
        filter_dict = data['filters'].get(verticespec['joining_value'], {})
        # Also find out whether the path is used in a filter or a project and, if so,
        # instruct the recursive function to build the path on the fly.
        # The default is False, because it's super expensive
        expand_path = (data['filters'][edge_tag].get('path', None) is not None) or any(
            'path' in d.keys() for d in data['project'][edge_tag]
        )
        join = join_func(
            join_tag,
            join_to,
            isouterjoin=verticespec.get('outerjoin'),  # type: ignore
            filter_dict=filter_dict,
            expand_path=expand_path,
        )
        join.edge_tag = edge_tag
        joins.append(join)
    return joins


def generate_projections(
    data: QueryDictType, aliases: Dict[str, Optional[AliasedClass]], outer_to_inner_schema, get_projectable_entity
):
    """Generate the projections for the query."""
    # mapping of tag -> field -> projection_index
    tag_to_projected_fields: Dict[str, Dict[str, int]] = {}
    projections: List[Tuple[Union[AliasedClass, ColumnElement], bool]] = []

    QUERYBUILD_LOGGER.debug('projections data: %s', data['project'])

    if not any(data['project'].values()):
        # If user has not set projection,
        # I will simply project the last item specified!
        # Don't change, path traversal querying relies on this behavior!
        projections += _create_projections(
            data['path'][-1]['tag'],
            aliases,
            len(projections),
            [{'*': {}}],
            get_projectable_entity,
            tag_to_projected_fields,
            outer_to_inner_schema,
        )
    else:
        for vertex in data['path']:
            project_dict = data['project'].get(vertex['tag'], [])
            projections += _create_projections(
                vertex['tag'],
                aliases,
                len(projections),
                project_dict,
                get_projectable_entity,
                tag_to_projected_fields,
                outer_to_inner_schema,
            )

        for vertex in data['path'][1:]:
            edge_tag = vertex.get('edge_tag', None)

            QUERYBUILD_LOGGER.debug(
                'Checking projections for edges: This is edge %s from %s, %s of %s',
                edge_tag,
                vertex.get('tag'),
                vertex.get('joining_keyword'),
                vertex.get('joining_value'),
            )
            if edge_tag is not None:
                project_dict = data['project'].get(edge_tag, [])
                projections += _create_projections(
                    edge_tag,
                    aliases,
                    len(projections),
                    project_dict,
                    get_projectable_entity,
                    tag_to_projected_fields,
                    outer_to_inner_schema,
                )

    # check the consistency of projections
    projection_index_to_field = {
        index_in_sql_result: attrkey
        for _, projected_entities_dict in tag_to_projected_fields.items()
        for attrkey, index_in_sql_result in projected_entities_dict.items()
    }
    if len(projections) > len(projection_index_to_field):
        raise ValueError('You are projecting the same key multiple times within the same node')
    if not projection_index_to_field:
        raise ValueError('No projections requested')
    return projections, tag_to_projected_fields


def _create_projections(
    tag: str,
    aliases: Dict[str, Optional[AliasedClass]],
    projection_count: int,
    project_dict: List[Dict[str, dict]],
    get_projectable_entity,
    tag_to_projected_fields,
    outer_to_inner_schema,
) -> List[Tuple[Union[AliasedClass, ColumnElement], bool]]:
    """Build the projections for a given tag.

    :param tag: the tag of the node for which to build the projections
    :param projection_count: the number of previous projections
    :param items_to_project: the list of projections to build, if None, use the projections specified in the query

    """
    # Return here if there is nothing to project, reduces number of key in return dictionary
    QUERYBUILD_LOGGER.debug('projection for %s: %s', tag, project_dict)
    if not project_dict:
        return []

    alias = aliases.get(tag)
    if alias is None:
        raise ValueError(f'No alias found for tag {tag}')

    tag_to_projected_fields[tag] = {}

    projections = []

    for projectable_spec in project_dict:
        for projectable_entity_name, extraspec in projectable_spec.items():
            property_names = []
            if projectable_entity_name == '**':
                # Need to expand
                property_names.extend(modify_expansions(alias, get_column_names(alias), outer_to_inner_schema))
            else:
                property_names.extend(modify_expansions(alias, [projectable_entity_name], outer_to_inner_schema))

            for property_name in property_names:
                projections.append(_get_projection(alias, property_name, get_projectable_entity, **extraspec))
                tag_to_projected_fields[tag][property_name] = projection_count
                projection_count += 1

    return projections


def _get_projection(
    alias: AliasedClass,
    projectable_entity_name: str,
    get_projectable_entity,
    cast: Optional[str] = None,
    func: Optional[str] = None,
    **_kw: Any,
) -> Tuple[Union[AliasedClass, ColumnElement], bool]:
    """:param alias: An alias for an ormclass
    :param projectable_entity_name:
        User specification of what to project.
        Appends to query's entities what the user wants to project
        (have returned by the query)
    :param cast: Cast the value to a different type
    :param func: Apply a function to the projection

    :return: The projection
    """
    column_name = projectable_entity_name.split('.')[0]
    attr_key = projectable_entity_name.split('.')[1:]

    if column_name == '*':
        if func is not None:
            raise ValueError(
                'Very sorry, but functions on the aliased class\n'
                "(You specified '*')\n"
                'will not work!\n'
                "I suggest you apply functions on a column, e.g. ('id')\n"
            )
        return alias, True

    entity_to_project = get_projectable_entity(alias, column_name, attr_key, cast=cast)
    if func is None:
        pass
    elif func == 'max':
        entity_to_project = sa_func.max(entity_to_project)
    elif func == 'min':
        entity_to_project = sa_func.max(entity_to_project)
    elif func == 'count':
        entity_to_project = sa_func.count(entity_to_project)
    else:
        raise ValueError(f'\nInvalid function specification {func}')

    return entity_to_project, False



"""Parsers for DiffCalculation of plugin tutorial.

Register parsers via the "aiida.parsers" entry point in the pyproject.toml file.
"""

# mypy: disable_error_code=call-overload
# START PARSER HEAD
from aiida.engine import ExitCode
from aiida.orm import SinglefileData
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory

DiffCalculation = CalculationFactory('diff-tutorial')


class DiffParser(Parser):
    # END PARSER HEAD
    """Parser class for DiffCalculation."""

    def parse(self, **kwargs):
        """Parse outputs, store results in database.

        :returns: non-zero exit code, if parsing fails
        """
        output_filename = self.node.get_option('output_filename')

        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [output_filename]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error(f"Found files '{files_retrieved}', expected to find '{files_expected}'")
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        # add output file
        self.logger.info(f"Parsing '{output_filename}'")
        with self.retrieved.open(output_filename, 'rb') as handle:
            output_node = SinglefileData(file=handle)
        self.out('diff', output_node)

        return ExitCode(0)


class DiffParserSimple(Parser):
    """Simple Parser class for DiffCalculation."""

    def parse(self, **kwargs):
        """Parse outputs, store results in database."""
        output_filename = self.node.get_option('output_filename')

        # add output file
        self.logger.info(f"Parsing '{output_filename}'")
        with self.retrieved.open(output_filename, 'rb') as handle:
            output_node = SinglefileData(file=handle)
        self.out('diff', output_node)

        return ExitCode(0)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test running a :class:`~aiida.orm.nodes.data.codes.containerized.ContainerizedCode` code."""

from aiida import orm
from aiida.engine import run_get_node


def test_add_singularity():
    """Test installed containerized code by add plugin"""
    builder = orm.load_code('add-singularity@localhost').get_builder()
    builder.x = orm.Int(4)
    builder.y = orm.Int(6)
    builder.metadata.options.resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}

    results, node = run_get_node(builder)

    assert node.is_finished_ok
    assert 'sum' in results
    assert 'remote_folder' in results
    assert 'retrieved' in results
    assert results['sum'].value == 10


def main():
    test_add_singularity()


if __name__ == '__main__':
    main()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with all the internals that make up the engine of `aiida-core`."""

# AUTO-GENERATED

# fmt: off

from .daemon import *
from .exceptions import *
from .launch import *
from .persistence import *
from .processes import *
from .runners import *
from .utils import *

__all__ = (
    'AiiDAPersister',
    'Awaitable',
    'AwaitableAction',
    'AwaitableTarget',
    'BaseRestartWorkChain',
    'CalcJob',
    'CalcJobImporter',
    'CalcJobOutputPort',
    'CalcJobProcessSpec',
    'DaemonClient',
    'ExitCode',
    'ExitCodesNamespace',
    'FunctionProcess',
    'InputPort',
    'InterruptableFuture',
    'JobManager',
    'JobsList',
    'ObjectLoader',
    'OutputPort',
    'PORT_NAMESPACE_SEPARATOR',
    'PastException',
    'PortNamespace',
    'Process',
    'ProcessBuilder',
    'ProcessBuilderNamespace',
    'ProcessFuture',
    'ProcessHandlerReport',
    'ProcessSpec',
    'ProcessState',
    'Runner',
    'ToContext',
    'WithNonDb',
    'WithSerialize',
    'WorkChain',
    'append_',
    'assign_',
    'await_processes',
    'calcfunction',
    'construct_awaitable',
    'get_daemon_client',
    'get_object_loader',
    'if_',
    'interruptable_task',
    'is_process_function',
    'process_handler',
    'return_',
    'run',
    'run_get_node',
    'run_get_pk',
    'submit',
    'while_',
    'workfunction',
)

# fmt: on



"""Defaults related to RabbitMQ."""

from __future__ import annotations

import os
import typing as t

from aiida.common.extendeddicts import AttributeDict
from aiida.common.log import AIIDA_LOGGER

__all__ = ('BROKER_DEFAULTS',)

LOGGER = AIIDA_LOGGER.getChild('brokers.rabbitmq.defaults')

LAUNCH_QUEUE = 'process.queue'
MESSAGE_EXCHANGE = 'messages'
TASK_EXCHANGE = 'tasks'

BROKER_DEFAULTS = AttributeDict(
    {
        'protocol': 'amqp',
        'username': 'guest',
        'password': 'guest',
        'host': '127.0.0.1',
        'port': 5672,
        'virtual_host': '',
        'heartbeat': 600,
    }
)


def detect_rabbitmq_config(
    protocol: str | None = None,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    virtual_host: str | None = None,
) -> dict[str, t.Any]:
    """Try to connect to a RabbitMQ server with the default connection parameters.

    :raises ConnectionError: If the connection failed with the provided connection parameters
    :returns: The connection parameters if the RabbitMQ server was successfully connected to, or ``None`` otherwise.
    """
    from kiwipy.rmq.threadcomms import connect

    connection_params = {
        'protocol': protocol or os.getenv('AIIDA_BROKER_PROTOCOL', BROKER_DEFAULTS['protocol']),
        'username': username or os.getenv('AIIDA_BROKER_USERNAME', BROKER_DEFAULTS['username']),
        'password': password or os.getenv('AIIDA_BROKER_PASSWORD', BROKER_DEFAULTS['password']),
        'host': host or os.getenv('AIIDA_BROKER_HOST', BROKER_DEFAULTS['host']),
        'port': port or int(os.getenv('AIIDA_BROKER_PORT', BROKER_DEFAULTS['port'])),
        'virtual_host': virtual_host or os.getenv('AIIDA_BROKER_VIRTUAL_HOST', BROKER_DEFAULTS['virtual_host']),
    }

    LOGGER.info(f'Attempting to connect to RabbitMQ with parameters: {connection_params}')

    try:
        connect(connection_params=connection_params)
    except ConnectionError:
        raise ConnectionError(f'Failed to connect with following connection parameters: {connection_params}')

    # The profile configuration expects the keys of the broker config to be prefixed with ``broker_``.
    return {f'broker_{key}': value for key, value in connection_params.items()}



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This allows to manage showfunctionality to all data types."""

import pathlib

from aiida.cmdline.utils import echo
from aiida.common.exceptions import MultipleObjectsError


def has_executable(exec_name):
    """:return: True if executable can be found in PATH, False otherwise."""
    import shutil

    return shutil.which(exec_name) is not None


def _show_jmol(exec_name, trajectory_list, **_kwargs):
    """Plugin for jmol"""
    import subprocess
    import tempfile

    if not has_executable(exec_name):
        echo.echo_critical(f"No executable '{exec_name}' found. Add to the path, or try with an absolute path.")

    with tempfile.NamedTemporaryFile(mode='w+b') as handle:
        for trajectory in trajectory_list:
            handle.write(trajectory._exportcontent('cif')[0])
        handle.flush()

        try:
            subprocess.check_output([exec_name, handle.name])
        except subprocess.CalledProcessError:
            # The program died: just print a message
            echo.echo_error(f'the call to {exec_name} ended with an error.')


def _show_xcrysden(exec_name, trajectory_list, **_kwargs):
    """Plugin for xcrysden"""
    import subprocess
    import tempfile

    if len(trajectory_list) > 1:
        raise MultipleObjectsError('Visualization of multiple trajectories is not implemented')
    obj = trajectory_list[0]

    if not has_executable(exec_name):
        echo.echo_critical(f"No executable '{exec_name}' found.")

    with tempfile.NamedTemporaryFile(mode='w+b', suffix='.xsf') as tmpf:
        tmpf.write(obj._exportcontent('xsf')[0])
        tmpf.flush()

        try:
            subprocess.check_output([exec_name, '--xsf', tmpf.name])
        except subprocess.CalledProcessError:
            # The program died: just print a message
            echo.echo_error(f'the call to {exec_name} ended with an error.')


def _show_mpl_pos(exec_name, trajectory_list, **kwargs):
    """Produces a matplotlib plot of the trajectory"""
    for traj in trajectory_list:
        traj.show_mpl_pos(**kwargs)


def _show_mpl_heatmap(exec_name, trajectory_list, **kwargs):
    """Produces a matplotlib plot of the trajectory"""
    for traj in trajectory_list:
        traj.show_mpl_heatmap(**kwargs)


def _show_ase(exec_name, structure_list):
    """Plugin to show the structure with the ASE visualizer"""
    try:
        from ase.visualize import view

        for structure in structure_list:
            view(structure.get_ase())
    except ImportError:
        raise


def _show_vesta(exec_name, structure_list):
    """Plugin for VESTA
    This VESTA plugin was added by Yue-Wen FANG and Abel Carreras
    at Kyoto University in the group of Prof. Isao Tanaka's lab

    """
    import subprocess
    import tempfile

    with tempfile.NamedTemporaryFile(mode='w+b', suffix='.cif') as tmpf:
        for structure in structure_list:
            tmpf.write(structure._exportcontent('cif')[0])
        tmpf.flush()

        try:
            subprocess.check_output([exec_name, tmpf.name])
        except subprocess.CalledProcessError:
            # The program died: just print a message
            echo.echo_error(f'the call to {exec_name} ended with an error.')
        except OSError as err:
            if err.errno == 2:
                echo.echo_critical(f"No executable '{exec_name}' found. Add to the path, or try with an absolute path.")
            else:
                raise


def _show_vmd(exec_name, structure_list):
    """Plugin for vmd"""
    import subprocess
    import tempfile

    if len(structure_list) > 1:
        raise MultipleObjectsError('Visualization of multiple objects is not implemented')
    structure = structure_list[0]

    with tempfile.NamedTemporaryFile(mode='w+b', suffix='.xsf') as tmpf:
        tmpf.write(structure._exportcontent('xsf')[0])
        tmpf.flush()

        try:
            subprocess.check_output([exec_name, tmpf.name])
        except subprocess.CalledProcessError:
            # The program died: just print a message
            echo.echo_error(f'the call to {exec_name} ended with an error.')
        except OSError as err:
            if err.errno == 2:
                echo.echo_critical(f"No executable '{exec_name}' found. Add to the path, or try with an absolute path.")
            else:
                raise


def _show_xmgrace(exec_name, list_bands):
    """Plugin for showing the bands with the XMGrace plotting software."""
    import subprocess
    import sys
    import tempfile

    from aiida.orm.nodes.data.array.bands import MAX_NUM_AGR_COLORS

    list_files = []
    current_band_number = 0

    with tempfile.TemporaryDirectory() as tmpdir:
        dirpath = pathlib.Path(tmpdir)

        for iband, bnds in enumerate(list_bands):
            # extract number of bands
            nbnds = bnds.get_bands().shape[1]
            text, _ = bnds._exportcontent(
                'agr', setnumber_offset=current_band_number, color_number=(iband + 1 % MAX_NUM_AGR_COLORS)
            )
            # write a tempfile
            filepath = dirpath / f'{iband}.agr'
            filepath.write_bytes(text)
            list_files.append(str(filepath))
            # update the number of bands already plotted
            current_band_number += nbnds

        try:
            subprocess.check_output([exec_name] + [str(filepath) for filepath in list_files])
        except subprocess.CalledProcessError:
            print(f'Note: the call to {exec_name} ended with an error.')
        except OSError as err:
            if err.errno == 2:
                print(f"No executable '{exec_name}' found. Add to the path, or try with an absolute path.")
                sys.exit(1)
            else:
                raise



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Delete trajectory symbols array from the repository and the reference in the attributes

Note, this is similar to the django migration django_0027

Revision ID: ce56d84bcc35
Revises: 12536798d4d3
Create Date: 2019-01-21 15:35:07.280805

"""

from alembic import op
from sqlalchemy import Integer, String
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.sql import column, select, table, text

from aiida.storage.psql_dos.backend import get_filepath_container
from aiida.storage.psql_dos.migrations.utils import utils

# revision identifiers, used by Alembic.
revision = 'ce56d84bcc35'
down_revision = '12536798d4d3'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    connection = op.get_bind()
    repo_path = get_filepath_container(op.get_context().opts['aiida_profile']).parent

    DbNode = table(  # noqa: N806
        'db_dbnode',
        column('id', Integer),
        column('uuid', UUID),
        column('type', String),
        column('attributes', JSONB),
    )

    nodes = connection.execute(
        select(DbNode.c.id, DbNode.c.uuid).where(
            DbNode.c.type == op.inline_literal('node.data.array.trajectory.TrajectoryData.')
        )
    ).fetchall()

    for pk, uuid in nodes:
        connection.execute(
            text(f"""UPDATE db_dbnode SET attributes = attributes #- '{{array|symbols}}' WHERE id = {pk}""")
        )
        utils.delete_numpy_array_from_repository(repo_path, uuid, 'symbols')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of ce56d84bcc35.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This allows to hook-up the AiiDA built-in RESTful API.
Main advantage of doing this by means of a verdi command is that different
profiles can be selected at hook-up (-p flag).
"""

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params.options import DEBUG, HOSTNAME, PORT
from aiida.restapi.common import config


@verdi.command('restapi')
@HOSTNAME(default=config.CLI_DEFAULTS['HOST_NAME'])
@PORT(default=config.CLI_DEFAULTS['PORT'])
@click.option(
    '-c',
    '--config-dir',
    type=click.Path(exists=True),
    default=config.CLI_DEFAULTS['CONFIG_DIR'],
    help='Path to the configuration directory',
)
@DEBUG(default=config.APP_CONFIG['DEBUG'])
@click.option(
    '--wsgi-profile',
    is_flag=True,
    default=config.CLI_DEFAULTS['WSGI_PROFILE'],
    help='Whether to enable WSGI profiler middleware for finding bottlenecks',
)
@click.option(
    '--posting/--no-posting',
    'posting',
    is_flag=True,
    default=config.CLI_DEFAULTS['POSTING'],
    help='Enable POST endpoints (currently only /querybuilder).',
    hidden=True,
)
@click.pass_context
def restapi(ctx, hostname, port, config_dir, debug, wsgi_profile, posting):
    """Run the AiiDA REST API server.

    Example Usage:

        verdi -p <profile_name> restapi --hostname 127.0.0.5 --port 6789
    """
    from aiida.restapi.run_api import run_api

    # Invoke the runner
    try:
        run_api(
            profile=ctx.obj['profile'].name,
            hostname=hostname,
            port=port,
            config=config_dir,
            debug=debug,
            wsgi_profile=wsgi_profile,
            posting=posting,
        )
    except ImportError as exc:
        raise ImportError(
            'Failed to import modules required for the REST API. '
            'You may need to install the `rest` extra, e.g. via `pip install aiida-core[rest]`.'
        ) from exc



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""An IPython extension that provides a magic command to load
basic aiida commands.

This makes it much easier to start.

Produces output in:

* Plaintext (IPython [qt]console)
* HTML (IPython notebook, ``nbconvert --to html``, ``--to slides``)
* JSON (IPython notebook ``.ipynb`` files)
* LaTeX (e.g. ``ipython nbconvert example.ipynb --to LaTeX --post PDF``)


Notes on how to load it at start:
https://ipython.org/ipython-doc/3/config/intro.html

Usage
======

.. sourcecode:: ipython

   In [1]: %load_ext aiida_magic

   In [2]: %aiida
"""

import json
import shlex

from IPython import get_ipython, version_info  # type: ignore[attr-defined]
from IPython.core import magic


def add_to_ns(local_ns, name, obj):
    """Add a new variable with name ``name`` and value ``obj`` to the
    namespace ``local_ns``, optionally showing a warning if we are
    hiding an existing variable.

    .. todo:: implement the warning.

    Example::

        # assuming that local_ns is a dictionary, e.g. from locals()
        import sys
        add_to_ns(local_ns, 'sys', sys)
    """
    if name in local_ns:
        pass
    local_ns[name] = obj


@magic.magics_class
class AiiDALoaderMagics(magic.Magics):
    """AiiDA magic loader."""

    @magic.needs_local_scope
    @magic.line_magic
    def verdi(self, line='', local_ns=None):
        """Run the AiiDA command line tool, using the currently loaded configuration and profile.

        Invoking ``verdi`` normally through the command line follows a different code path, compared to calling it
        directly from within an active Python interpreter. Some of those code paths we actually need here, and others
        can actually cause problems:

        * The ``VerdiCommandGroup`` group ensures that the context ``obj`` is set with the loaded ``Config`` and
          ``Profile``, but this is not done in this manual call, so we have to built it ourselves and pass it in with
          the ``obj`` keyword.
        * We cannot call the ``aiida.cmdline.commands.cmd_verdi.verdi`` command directly, as that will invoke the
          ``aiida.cmdline.parameters.types.profile.ProfileParamType`` parameter used for the ``-p/--profile`` option
          that is defined on the ``verdi`` command. This will attempt to load the default profile, but here a profile
          has actually already been loaded. In principle, reloading a profile should not be a problem, but this magic
          is often used in demo notebooks where a temporary profile was created and loaded, which is not properly added
          to the config file (since it is temporary) and so loading the profile would fail. The solution is to not call
          the top-level ``verdi`` command, but the subcommand, which is the first parameter in the command line
          arguments defined by the ``line`` argument.

        """
        from click import Context

        from aiida.cmdline.commands.cmd_verdi import verdi
        from aiida.common import AttributeDict
        from aiida.manage import get_config, get_profile

        config = get_config()
        profile = get_profile()

        # The ``line`` should be of the form ``process list -a -p1``, i.e., a ``verdi`` subcommand with some optional
        # parameters. Validate that the first entry is indeed a subcommand and not the flag to specify the profile which
        # is not supported in this magic method.
        cmdline_arguments = shlex.split(line)
        command_name = cmdline_arguments[0]

        if command_name in ('-p', '--profile'):
            raise ValueError(
                'The `-p/--profile` option is not supported for the `%verdi` magic operator. It will use the currently '
                f'loaded profile `{profile}`'
            )

        # Construct the subcommand that will be executed, thereby circumventing the profile option of ``verdi`` itself.
        # If the caller specified a subcommand that doesn't exist, the following will raise an exception.
        context = Context(verdi)
        command = verdi.get_command(context, command_name)

        if command is None:
            raise RuntimeError(f'command `{command_name}` not found.')

        return command(
            cmdline_arguments[1:],
            prog_name='%verdi',
            obj=AttributeDict({'config': config, 'profile': profile}),
            standalone_mode=False,
        )

    @magic.needs_local_scope
    @magic.line_magic
    def aiida(self, line='', local_ns=None):
        """Load AiiDA in ipython (checking if it was already loaded), and
        inserts in the namespace the main AiiDA classes (the same that are
        loaded in ``verdi shell``.

        Usage::

            %aiida [optional parameters]

        .. todo:: implement parameters, e.g. for the profile to load.
        """
        from aiida.cmdline.utils.shell import get_start_namespace
        from aiida.manage.configuration import load_profile

        self.is_warning = False
        lcontent = line.strip()
        if lcontent:
            profile = load_profile(lcontent)
        else:
            profile = load_profile()

        self.current_state = f'Loaded AiiDA DB environment - profile name: {profile.name}.'

        user_ns = get_start_namespace()
        for key, value in user_ns.items():
            add_to_ns(local_ns, key, value)

        return self

    def _repr_json_(self):
        """Output in JSON format."""
        obj = {'current_state': self.current_state}
        if version_info[0] >= 3:
            return obj

        return json.dumps(obj)

    def _repr_html_(self):
        """Output in HTML format."""
        html = '<p>'
        if self.is_warning:
            html += '<strong>'
        html += self.current_state
        if self.is_warning:
            html += '</strong>'
        html += '</p>'

        return html

    def _repr_latex_(self):
        """Output in LaTeX format."""
        if self.is_warning:
            latex = '\\emph{%s}\n' % self.current_state
        else:
            latex = f'{self.current_state}\n'

        return latex

    def _repr_pretty_(self, pretty_print, cycle):
        """Output in text format."""
        if self.is_warning:
            warning_str = '** '
        else:
            warning_str = ''
        text = f'{warning_str}{self.current_state}\n'

        pretty_print.text(text)


def load_ipython_extension(ipython):
    """Registers the %aiida IPython extension.

    .. deprecated:: v3.0.0
        Use :py:func:`~aiida.tools.ipython.ipython_magics.register_ipython_extension` instead.
    """
    register_ipython_extension(ipython)


def register_ipython_extension(ipython=None):
    """Registers the %aiida IPython extension.

    The %aiida IPython extension provides the same environment as the `verdi shell`.

    :param ipython: InteractiveShell instance. If omitted, the global InteractiveShell is used.

    """
    if ipython is None:
        ipython = get_ipython()

    ipython.register_magics(AiiDALoaderMagics)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA specific implementation of plumpy's ProcessSpec."""

from typing import Optional

import plumpy.process_spec

from aiida.orm import Dict

from .exit_code import ExitCode, ExitCodesNamespace
from .ports import CalcJobOutputPort, InputPort, PortNamespace

__all__ = ('ProcessSpec', 'CalcJobProcessSpec')


class ProcessSpec(plumpy.process_spec.ProcessSpec):
    """Default process spec for process classes defined in `aiida-core`.

    This sub class defines custom classes for input ports and port namespaces. It also adds support for the definition
    of exit codes and retrieving them subsequently.
    """

    METADATA_KEY: str = 'metadata'
    METADATA_OPTIONS_KEY: str = 'options'
    INPUT_PORT_TYPE = InputPort
    PORT_NAMESPACE_TYPE = PortNamespace

    def __init__(self) -> None:
        super().__init__()
        self._exit_codes = ExitCodesNamespace()

    @property
    def metadata_key(self) -> str:
        return self.METADATA_KEY

    @property
    def options_key(self) -> str:
        return self.METADATA_OPTIONS_KEY

    @property
    def exit_codes(self) -> ExitCodesNamespace:
        """Return the namespace of exit codes defined for this ProcessSpec

        :returns: ExitCodesNamespace of ExitCode named tuples
        """
        return self._exit_codes

    def exit_code(self, status: int, label: str, message: str, invalidates_cache: bool = False) -> None:
        """Add an exit code to the ProcessSpec

        :param status: the exit status integer
        :param label: a label by which the exit code can be addressed
        :param message: a more detailed description of the exit code
        :param invalidates_cache: when set to `True`, a process exiting
            with this exit code will not be considered for caching
        """
        if not isinstance(status, int):
            raise TypeError(f'status should be of integer type and not of {type(status)}')

        if status < 0:
            raise ValueError(f'status should be a positive integer, received {type(status)}')

        if not isinstance(label, str):
            raise TypeError(f'label should be of str type and not of {type(label)}')

        if not isinstance(message, str):
            raise TypeError(f'message should be of str type and not of {type(message)}')

        if not isinstance(invalidates_cache, bool):
            raise TypeError(f'invalidates_cache should be of type bool and not of {type(invalidates_cache)}')

        self._exit_codes[label] = ExitCode(status, message, invalidates_cache=invalidates_cache)

    # override return type to aiida's PortNamespace subclass

    @property
    def ports(self) -> PortNamespace:
        return super().ports  # type: ignore[return-value]

    @property
    def inputs(self) -> PortNamespace:
        return super().inputs  # type: ignore[return-value]

    @property
    def outputs(self) -> PortNamespace:
        return super().outputs  # type: ignore[return-value]


class CalcJobProcessSpec(ProcessSpec):
    """Process spec intended for the `CalcJob` process class."""

    OUTPUT_PORT_TYPE = CalcJobOutputPort

    def __init__(self) -> None:
        super().__init__()
        self._default_output_node: Optional[str] = None

    @property
    def default_output_node(self) -> Optional[str]:
        return self._default_output_node

    @default_output_node.setter
    def default_output_node(self, port_name: str) -> None:
        if port_name not in self.outputs:
            raise ValueError(f'{port_name} is not a registered output port')

        valid_type_port = self.outputs[port_name].valid_type
        valid_type_required = Dict

        if valid_type_port is not valid_type_required:
            raise ValueError(
                f'the valid type of a default output has to be a {valid_type_required} but it is {valid_type_port}'
            )

        self._default_output_node = port_name



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Top level functions that can be used to launch a Process."""

from __future__ import annotations

import time
import typing as t

from aiida.common import InvalidOperation
from aiida.common.lang import type_check
from aiida.common.log import AIIDA_LOGGER
from aiida.manage import manager
from aiida.orm import ProcessNode

from .processes.builder import ProcessBuilder
from .processes.functions import FunctionProcess
from .processes.process import Process
from .runners import ResultAndPk
from .utils import instantiate_process, is_process_scoped, prepare_inputs

__all__ = ('run', 'run_get_pk', 'run_get_node', 'submit', 'await_processes')

TYPE_RUN_PROCESS = t.Union[Process, t.Type[Process], ProcessBuilder]
# run can also be process function, but it is not clear what type this should be
TYPE_SUBMIT_PROCESS = t.Union[Process, t.Type[Process], ProcessBuilder]
LOGGER = AIIDA_LOGGER.getChild('engine.launch')


def run(process: TYPE_RUN_PROCESS, inputs: dict[str, t.Any] | None = None, **kwargs: t.Any) -> dict[str, t.Any]:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class or process function to run
    :param inputs: the inputs to be passed to the process
    :return: the outputs of the process
    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run(process, inputs, **kwargs)


def run_get_node(
    process: TYPE_RUN_PROCESS, inputs: dict[str, t.Any] | None = None, **kwargs: t.Any
) -> tuple[dict[str, t.Any], ProcessNode]:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class, instance, builder or function to run
    :param inputs: the inputs to be passed to the process
    :return: tuple of the outputs of the process and the process node
    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run_get_node(process, inputs, **kwargs)


def run_get_pk(process: TYPE_RUN_PROCESS, inputs: dict[str, t.Any] | None = None, **kwargs: t.Any) -> ResultAndPk:
    """Run the process with the supplied inputs in a local runner that will block until the process is completed.

    :param process: the process class, instance, builder or function to run
    :param inputs: the inputs to be passed to the process
    :return: tuple of the outputs of the process and process node pk
    """
    if isinstance(process, Process):
        runner = process.runner
    else:
        runner = manager.get_manager().get_runner()

    return runner.run_get_pk(process, inputs, **kwargs)


def submit(
    process: TYPE_SUBMIT_PROCESS,
    inputs: dict[str, t.Any] | None = None,
    *,
    wait: bool = False,
    wait_interval: int = 5,
    **kwargs: t.Any,
) -> ProcessNode:
    """Submit the process with the supplied inputs to the daemon immediately returning control to the interpreter.

    .. warning: this should not be used within another process. Instead, there one should use the ``submit`` method of
        the wrapping process itself, i.e. use ``self.submit``.

    .. warning: submission of processes requires ``store_provenance=True``.

    :param process: the process class, instance or builder to submit
    :param inputs: the input dictionary to be passed to the process
    :param wait: when set to ``True``, the submission will be blocking and wait for the process to complete at which
        point the function returns the calculation node.
    :param wait_interval: the number of seconds to wait between checking the state of the process when ``wait=True``.
    :param kwargs: inputs to be passed to the process. This is an alternative to the positional ``inputs`` argument.
    :return: the calculation node of the process
    """
    inputs = prepare_inputs(inputs, **kwargs)

    # Submitting from within another process requires ``self.submit``` unless it is a work function, in which case the
    # current process in the scope should be an instance of ``FunctionProcess``.
    if is_process_scoped() and not isinstance(Process.current(), FunctionProcess):
        raise InvalidOperation('Cannot use top-level `submit` from within another process, use `self.submit` instead')

    runner = manager.get_manager().get_runner()

    if runner.controller is None:
        raise InvalidOperation(
            'Cannot submit because the runner does not have a process controller, probably because the profile does '
            'not define a broker like RabbitMQ. If a RabbitMQ server is available, the profile can be configured to '
            'use it with `verdi profile configure-rabbitmq`. Otherwise, use :meth:`aiida.engine.launch.run` instead to '
            'run the process in the local Python interpreter instead of submitting it to the daemon.'
        )

    assert runner.persister is not None, 'runner does not have a persister'

    process_inited = instantiate_process(runner, process, **inputs)

    # If a dry run is requested, simply forward to `run`, because it is not compatible with `submit`. We choose for this
    # instead of raising, because in this way the user does not have to change the launcher when testing. The same goes
    # for if `remote_folder` is present in the inputs, which means we are importing an already completed calculation.
    if process_inited.metadata.get('dry_run', False) or 'remote_folder' in inputs:
        _, node = run_get_node(process_inited)
        return node

    if not process_inited.metadata.store_provenance:
        raise InvalidOperation('cannot submit a process with `store_provenance=False`')

    runner.persister.save_checkpoint(process_inited)
    process_inited.close()

    # Do not wait for the future's result, because in the case of a single worker this would cock-block itself
    runner.controller.continue_process(process_inited.pid, nowait=False, no_reply=True)
    node = process_inited.node

    if not wait:
        return node

    while not node.is_terminated:
        LOGGER.report(
            f'Process<{node.pk}> has not yet terminated, current state is `{node.process_state}`. '
            f'Waiting for {wait_interval} seconds.'
        )
        time.sleep(wait_interval)

    return node


def await_processes(nodes: t.Sequence[ProcessNode], wait_interval: int = 1) -> None:
    """Run a loop until all processes are terminated.

    :param nodes: Sequence of nodes that represent the processes to await.
    :param wait_interval: The interval between each iteration of checking the status of all processes.
    """
    type_check(nodes, (list, tuple))

    if any(not isinstance(node, ProcessNode) for node in nodes):
        raise TypeError(f'`nodes` should be a list of `ProcessNode`s but got: {nodes}')

    start_time = time.time()
    terminated = False

    while not terminated:
        running = [not node.is_terminated for node in nodes]
        terminated = not any(running)
        seconds_passed = time.time() - start_time
        LOGGER.report(f'{running.count(False)} out of {len(nodes)} processes terminated. [{round(seconds_passed)} s]')
        time.sleep(wait_interval)


# Allow one to also use run.get_node and run.get_pk as a shortcut, without having to import the functions themselves
run.get_node = run_get_node  # type: ignore[attr-defined]
run.get_pk = run_get_pk  # type: ignore[attr-defined]



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub classes for data and processes."""

# AUTO-GENERATED

# fmt: off

from .attributes import *
from .data import *
from .node import *
from .process import *
from .repository import *

__all__ = (
    'AbstractCode',
    'ArrayData',
    'BandsData',
    'BaseType',
    'Bool',
    'CalcFunctionNode',
    'CalcJobNode',
    'CalculationNode',
    'CifData',
    'Code',
    'ContainerizedCode',
    'Data',
    'Dict',
    'EnumData',
    'Float',
    'FolderData',
    'InstalledCode',
    'Int',
    'JsonableData',
    'Kind',
    'KpointsData',
    'List',
    'Node',
    'NodeAttributes',
    'NodeRepository',
    'NumericType',
    'OrbitalData',
    'PortableCode',
    'ProcessNode',
    'ProjectionData',
    'RemoteData',
    'RemoteStashData',
    'RemoteStashFolderData',
    'SinglefileData',
    'Site',
    'Str',
    'StructureData',
    'TrajectoryData',
    'UpfData',
    'WorkChainNode',
    'WorkFunctionNode',
    'WorkflowNode',
    'XyData',
    'cif_from_ase',
    'find_bandgap',
    'has_pycifrw',
    'pycifrw_from_cif',
    'to_aiida_type',
)

# fmt: on



"""A sub class of ``plumpy.ProcessLauncher`` to launch a ``Process``."""

import asyncio
import logging
import traceback

import kiwipy
import plumpy

LOGGER = logging.getLogger(__name__)


class ProcessLauncher(plumpy.ProcessLauncher):
    """A sub class of :class:`plumpy.ProcessLauncher` to launch a ``Process``.

    It overrides the _continue method to make sure the node corresponding to the task can be loaded and
    that if it is already marked as terminated, it is not continued but the future is reconstructed and returned
    """

    @staticmethod
    def handle_continue_exception(node, exception, message):
        """Handle exception raised in `_continue` call.

        If the process state of the node has not yet been put to excepted, the exception was raised before the process
        instance could be reconstructed, for example when the process class could not be loaded, thereby circumventing
        the exception handling of the state machine. Raising this exception will then acknowledge the process task with
        RabbitMQ leaving an uncleaned node in the `CREATED` state for ever. Therefore we have to perform the node
        cleaning manually.

        :param exception: the exception object
        :param message: string message to use for the log message
        """
        from aiida.engine import ProcessState

        if not node.is_excepted and not node.is_sealed:
            node.logger.exception(message)
            node.set_exception(''.join(traceback.format_exception(type(exception), exception, None)).rstrip())
            node.set_process_state(ProcessState.EXCEPTED)
            node.seal()

    async def _continue(self, communicator, pid, nowait, tag=None):
        """Continue the task.

        Note that the task may already have been completed, as indicated from the corresponding the node, in which
        case it is not continued, but the corresponding future is reconstructed and returned. This scenario may
        occur when the Process was already completed by another worker that however failed to send the acknowledgment.

        :param communicator: the communicator that called this method
        :param pid: the pid of the process to continue
        :param nowait: if True don't wait for the process to finish, just return the pid, otherwise wait and
            return the results
        :param tag: the tag of the checkpoint to continue from
        """
        from aiida.common import exceptions
        from aiida.engine.exceptions import PastException
        from aiida.orm import Data, load_node
        from aiida.orm.utils import serialize

        try:
            node = load_node(pk=pid)
        except (exceptions.MultipleObjectsError, exceptions.NotExistent):
            # In this case, the process node corresponding to the process id, cannot be resolved uniquely or does not
            # exist. The latter being the most common case, where someone deleted the node, before the process was
            # properly terminated. Since the node is never coming back and so the process will never be able to continue
            # we raise `Return` instead of `TaskRejected` because the latter would cause the task to be resent and start
            # to ping-pong between RabbitMQ and the daemon workers.
            LOGGER.exception('Cannot continue process<%d>', pid)
            return False

        if node.is_terminated:
            LOGGER.info('not continuing process<%d> which is already terminated with state %s', pid, node.process_state)

            future = kiwipy.Future()

            if node.is_finished:
                future.set_result(
                    {entry.link_label: entry.node for entry in node.base.links.get_outgoing(node_class=Data)}
                )
            elif node.is_excepted:
                future.set_exception(PastException(node.exception))
            elif node.is_killed:
                future.set_exception(plumpy.KilledError())

            return future.result()

        try:
            result = await super()._continue(communicator, pid, nowait, tag)
        except ImportError as exception:
            message = 'the class of the process could not be imported.'
            self.handle_continue_exception(node, exception, message)
            raise
        except kiwipy.DuplicateSubscriberIdentifier:
            # This happens when the current worker has already subscribed itself with this process identifier. The call
            # to ``_continue`` will call ``Process.init`` which will add RPC and broadcast subscribers. ``kiwipy`` and
            # ``aiormq`` further down keep track of processes that are already subscribed and if subscribed again, a
            # ``DuplicateSubscriberIdentifier`` is raised. Possible reasons for the worker receiving a process task that
            # it already has, include:
            #
            #  1. The user mistakenly recreates the corresponding task, thinking the original task was lost.
            #  2. RabbitMQ requeues the task because the daemon worker lost its connection or did not respond to the
            #     heartbeat in time, and the task is sent to the same worker once it regains connection.
            #
            # Here we assume that the existence of another subscriber indicates that the process is still being run by
            # this worker. We thus ignore the request to have the worker take it on again and acknowledge the current
            # task (`return False`). If our assumption was wrong and the original task was no longer being worked on,
            # the user can resubmit the task once the list of subscribers of the process has been cleared. Note: In the
            # second case we are deleting the *original* task, and once the worker finishes running the process there
            # won't be a task in RabbitMQ to acknowledge anymore. This, however, is silently ignored.
            #
            # Note: the exception is raised by ``kiwipy`` based on an internal cache it and ``aiormq`` keep of the
            # current subscribers. This means that this will only occur when the tasks is resent to the *same* daemon
            # worker. If another worker were to receive it, no exception would be raised as the check is client and not
            # server based.
            LOGGER.exception(
                'A subscriber with the process id<%d> already exists, which most likely means this worker is already '
                'working on it and this task was sent as a duplicate by mistake. Deleting the task now.',
                pid,
            )
            return False
        except asyncio.CancelledError:
            # note this is only required in python<=3.7,
            # where asyncio.CancelledError inherits from Exception
            raise
        except Exception as exception:
            message = 'failed to recreate the process instance in order to continue it.'
            self.handle_continue_exception(node, exception, message)
            raise

        # Ensure that the result is serialized such that communication thread won't have to do database operations
        try:
            serialized = serialize.serialize(result)
        except Exception:
            LOGGER.exception('failed to serialize the result for process<%d>', pid)
            raise

        return serialized



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi user` command."""

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.params.options.commands import setup as options_setup
from aiida.cmdline.utils import decorators, echo


@verdi.group('user')
def verdi_user():
    """Inspect and manage users."""


@verdi_user.command('list')
@decorators.with_dbenv()
def user_list():
    """Show a list of all users."""
    from aiida.orm import User

    table = []

    for user in sorted(User.collection.all(), key=lambda user: user.email):
        row = ['*' if user.is_default else '', user.email, user.first_name, user.last_name, user.institution]
        if user.is_default:
            table.append(list(map(echo.highlight_string, row)))
        else:
            table.append(row)

    echo.echo_tabulate(table, headers=['', 'Email', 'First name', 'Last name', 'Institution'])
    echo.echo('')

    if User.collection.get_default() is None:
        echo.echo_warning('No default user has been configured')
    else:
        echo.echo_report('The user highlighted and marked with a * is the default user.')


@verdi_user.command('configure')
@click.option(
    '--email',
    'user',
    prompt='User email',
    help='Email address that serves as the user name and a way to identify data created by it.',
    type=types.UserParamType(create=True),
    cls=options.interactive.InteractiveOption,
)
@options_setup.SETUP_USER_FIRST_NAME(contextual_default=lambda ctx: ctx.params['user'].first_name)
@options_setup.SETUP_USER_LAST_NAME(contextual_default=lambda ctx: ctx.params['user'].last_name)
@options_setup.SETUP_USER_INSTITUTION(contextual_default=lambda ctx: ctx.params['user'].institution)
@click.option(
    '--set-default',
    prompt='Set as default?',
    help='Set the user as the default user.',
    is_flag=True,
    cls=options.interactive.InteractiveOption,
    contextual_default=lambda ctx: ctx.params['user'].is_default,
)
@click.pass_context
@decorators.with_dbenv()
def user_configure(ctx, user, first_name, last_name, institution, set_default):
    """Configure a new or existing user.

    An e-mail address is used as the user name.
    """
    action = 'updated' if user.is_stored else 'created'
    user.first_name = first_name
    user.last_name = last_name
    user.institution = institution
    user.store()

    echo.echo_success(f'Successfully {action} `{user.email}`.')

    if set_default:
        ctx.invoke(user_set_default, user=user)


@verdi_user.command('set-default')
@arguments.USER()
@click.pass_context
@decorators.with_dbenv()
def user_set_default(ctx, user):
    """Set a user as the default user for the profile."""
    from aiida.manage import get_manager

    get_manager().set_default_user_email(ctx.obj.profile, user.email)
    echo.echo_success(f'Set `{user.email}` as the default user for profile `{ctx.obj.profile.name}.`')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi devel` commands."""

import sys

import click

from aiida import get_profile
from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options, types
from aiida.cmdline.utils import decorators, echo
from aiida.common import exceptions


@verdi.group('devel')
def verdi_devel():
    """Commands for developers."""


@verdi_devel.command('check-load-time')
def devel_check_load_time():
    """Check for common indicators that slowdown `verdi`.

    Check for environment properties that negatively affect the responsiveness of the `verdi` command line interface.
    Known pathways that increase load time:

        * the database environment is loaded when it doesn't need to be
        * Unexpected `aiida.*` modules are imported

    If either of these conditions are true, the command will raise a critical error
    """
    from aiida.manage import get_manager

    loaded_aiida_modules = [key for key in sys.modules if key.startswith('aiida.')]
    aiida_modules_str = '\n- '.join(sorted(loaded_aiida_modules))
    echo.echo_info(f'aiida modules loaded:\n- {aiida_modules_str}')

    manager = get_manager()

    if manager.profile_storage_loaded:
        echo.echo_critical('potential `verdi` speed problem: database backend is loaded.')

    allowed = ('aiida.brokers', 'aiida.cmdline', 'aiida.common', 'aiida.manage', 'aiida.plugins', 'aiida.restapi')
    for loaded in loaded_aiida_modules:
        if not any(loaded.startswith(mod) for mod in allowed):
            echo.echo_critical(
                f'potential `verdi` speed problem: `{loaded}` module is imported which is not in: {allowed}'
            )

    echo.echo_success('no issues detected')


@verdi_devel.command('check-undesired-imports')
def devel_check_undesired_imports():
    """Check that verdi does not import python modules it shouldn't.

    Note: The blacklist was taken from the list of packages in the 'atomic_tools' extra but can be extended.
    """
    loaded_modules = 0

    for modulename in [
        'asyncio',
        'requests',
        'plumpy',
        'disk_objectstore',
        'paramiko',
        'seekpath',
        'CifFile',
        'ase',
        'pymatgen',
        'spglib',
        'pymysql',
        'yaml',
    ]:
        if modulename in sys.modules:
            echo.echo_warning(f'Detected loaded module "{modulename}"')
            loaded_modules += 1

    if loaded_modules > 0:
        echo.echo_critical(f'Detected {loaded_modules} unwanted modules')
    echo.echo_success('no issues detected')


@verdi_devel.command('validate-plugins')
@decorators.with_dbenv()
def devel_validate_plugins():
    """Validate all plugins by checking they can be loaded."""
    from aiida.common.exceptions import EntryPointError
    from aiida.plugins.entry_point import validate_registered_entry_points

    try:
        validate_registered_entry_points()
    except EntryPointError as exception:
        echo.echo_critical(str(exception))

    echo.echo_success('all registered plugins could successfully loaded.')


@verdi_devel.command('run-sql')
@click.argument('sql', type=str)
def devel_run_sql(sql):
    """Run a raw SQL command on the profile database (only available for 'core.psql_dos' storage)."""
    from sqlalchemy import text

    from aiida.storage.psql_dos.utils import create_sqlalchemy_engine

    assert get_profile().storage_backend == 'core.psql_dos'
    with create_sqlalchemy_engine(get_profile().storage_config).connect() as connection:
        result = connection.execute(text(sql)).fetchall()

    if isinstance(result, (list, tuple)):
        for row in result:
            echo.echo(str(row))
    else:
        echo.echo(str(result))


@verdi_devel.command('play', hidden=True)
def devel_play():
    """Play the Aida triumphal march by Giuseppe Verdi."""
    import webbrowser

    webbrowser.open_new('http://upload.wikimedia.org/wikipedia/commons/3/32/Triumphal_March_from_Aida.ogg')


@verdi_devel.command('launch-add')
@options.CODE(type=types.CodeParamType(entry_point='core.arithmetic.add'))
@click.option('-d', '--daemon', is_flag=True, help='Submit to the daemon instead of running blockingly.')
@click.option('-s', '--sleep', type=int, help='Set the `sleep` input in seconds.')
def devel_launch_arithmetic_add(code, daemon, sleep):
    """Launch an ``ArithmeticAddCalculation``.

    Unless specified with the option ``--code``, a suitable ``Code`` is automatically setup. By default the command
    configures ``bash`` on the ``localhost``. If the localhost is not yet configured as a ``Computer``, that is also
    done automatically.
    """
    from shutil import which

    from aiida.engine import run, submit
    from aiida.orm import InstalledCode, Int, load_code

    default_calc_job_plugin = 'core.arithmetic.add'

    if not code:
        try:
            code = load_code('bash@localhost')
        except exceptions.NotExistent:
            localhost = prepare_localhost()
            code = InstalledCode(
                label='bash',
                computer=localhost,
                filepath_executable=which('bash'),
                default_calc_job_plugin=default_calc_job_plugin,
            ).store()
        else:
            assert code.default_calc_job_plugin == default_calc_job_plugin

    builder = code.get_builder()
    builder.x = Int(1)
    builder.y = Int(1)

    if sleep:
        builder.metadata.options.sleep = sleep

    if daemon:
        node = submit(builder)
        echo.echo_success(f'Submitted calculation `{node}`')
    else:
        _, node = run.get_node(builder)
        if node.is_finished_ok:
            echo.echo_success(f'ArithmeticAddCalculation<{node.pk}> finished successfully.')
        else:
            echo.echo_warning(f'ArithmeticAddCalculation<{node.pk}> did not finish successfully.')


def prepare_localhost():
    """Prepare and return the localhost as ``Computer``.

    If it doesn't already exist, the computer will be created, using ``core.local`` and ``core.direct`` as the entry
    points for the transport and scheduler type, respectively. In that case, the safe transport interval and the minimum
    job poll interval will both be set to 0 seconds in order to guarantee a throughput that is as fast as possible.

    :return: The localhost configured as a ``Computer``.
    """
    import tempfile

    from aiida.orm import Computer, load_computer

    try:
        computer = load_computer('localhost')
    except exceptions.NotExistent:
        echo.echo_warning('No `localhost` computer exists yet: creating and configuring the `localhost` computer.')
        computer = Computer(
            label='localhost',
            hostname='localhost',
            description='Localhost automatically created by `aiida.engine.launch_shell_job`',
            transport_type='core.local',
            scheduler_type='core.direct',
            workdir=tempfile.gettempdir(),
        ).store()
        computer.configure(safe_interval=0.0)
        computer.set_minimum_job_poll_interval(0.0)

    if not computer.is_configured:
        computer.configure()

    return computer



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions to operate on filesystem files."""

import hashlib


def md5_from_filelike(filelike, block_size_factor=128):
    """Create the hexdigested md5 checksum of the contents from a filelike object.

    :param filelike: the filelike object for whose contents to generate the md5 checksum
    :param block_size_factor: the file is read at chunks of size ``block_size_factor * md5.block_size``, where
        ``md5.block_size`` is the block_size used internally by the hashlib module.
    :returns: a string with the hexdigest md5.
    :raises: no checks are done on the filelike object, so it may raise OSError if it cannot be read from.
    """
    md5 = hashlib.md5()

    # I read 128 bytes at a time until it returns the empty string b''
    for chunk in iter(lambda: filelike.read(block_size_factor * md5.block_size), b''):
        md5.update(chunk)

    return md5.hexdigest()


def md5_file(filepath, block_size_factor=128):
    """Create the hexdigested md5 checksum of the contents from

    :param filepath: the filepath of the file for which we want the md5sum
    :param block_size_factor: the file is read at chunks of size ``block_size_factor * md5.block_size``, where
        ``md5.block_size`` is the block_size used internally by the hashlib module.
    :returns: a string with the hexdigest md5.
    :raises: No checks are done on the file, so if it doesn't exists it may
        raise OSError.
    """
    with open(filepath, 'rb', encoding=None) as handle:
        return md5_from_filelike(handle, block_size_factor=block_size_factor)


def sha1_file(filename, block_size_factor=128):
    """Open a file and return its sha1sum (hexdigested).

    :param filename: the filename of the file for which we want the sha1sum
    :param block_size_factor: the file is read at chunks of size
        ``block_size_factor * sha1.block_size``,
        where ``sha1.block_size`` is the block_size used internally by the
        hashlib module.

    :returns: a string with the hexdigest sha1.

    :raises: No checks are done on the file, so if it doesn't exists it may
        raise OSError.
    """
    sha1 = hashlib.sha1()
    with open(filename, 'rb', encoding=None) as fhandle:
        # I read 128 bytes at a time until it returns the empty string b''
        for chunk in iter(lambda: fhandle.read(block_size_factor * sha1.block_size), b''):
            sha1.update(chunk)
    return sha1.hexdigest()



"""Fixtures providing resources for storage plugins."""

from __future__ import annotations

import pathlib
import typing as t
from uuid import uuid4

import pytest
from pgtest.pgtest import PGTest

if t.TYPE_CHECKING:
    from pgtest.pgtest import PGTest


class PostgresCluster:
    def __init__(self):
        # We initialize the cluster lazily
        self.cluster = None

    def _create(self):
        try:
            self.cluster = PGTest()
        except OSError as e:
            raise RuntimeError('Could not initialize PostgreSQL cluster') from e

    def _close(self):
        if self.cluster is not None:
            self.cluster.close()

    def create_database(
        self,
        database_name: str | None = None,
        database_username: str | None = None,
        database_password: str | None = None,
    ) -> dict[str, str]:
        from aiida.manage.external.postgres import Postgres

        if self.cluster is None:
            self._create()

        postgres_config = {
            'database_engine': 'postgresql_psycopg2',
            'database_name': database_name or str(uuid4()),
            'database_username': database_username or 'guest',
            'database_password': database_password or 'guest',
        }

        postgres = Postgres(interactive=False, quiet=True, dbinfo=self.cluster.dsn)  # type: ignore[union-attr]
        if not postgres.dbuser_exists(postgres_config['database_username']):
            postgres.create_dbuser(
                postgres_config['database_username'], postgres_config['database_password'], 'CREATEDB'
            )
        postgres.create_db(postgres_config['database_username'], postgres_config['database_name'])

        postgres_config['database_hostname'] = postgres.host_for_psycopg2
        postgres_config['database_port'] = postgres.port_for_psycopg2

        return postgres_config


# TODO: Update docstring accordingly
@pytest.fixture(scope='session')
def postgres_cluster():
    """Create a temporary and isolated PostgreSQL cluster using ``pgtest`` and cleanup after the yield.

    :param database_name: Name of the database.
    :param database_username: Username to use for authentication.
    :param database_password: Password to use for authentication.
    :returns: Dictionary with parameters to connect to the PostgreSQL cluster.
    """

    cluster = PostgresCluster()
    yield cluster
    cluster._close()


@pytest.fixture(scope='session')
def config_psql_dos(
    tmp_path_factory: pytest.TempPathFactory,
    postgres_cluster: 'PGTest',
) -> t.Callable[[str | None, str | None, str | None], dict[str, t.Any]]:
    """Return a profile configuration for the :class:`~aiida.storage.psql_dos.backend.PsqlDosBackend`.

    The factory has the following signature to allow further configuring the database that is created:

    :param database_name: Name of the database to be created.
    :param database_username: Username to use for authentication.
    :param database_password: Password to use for authentication.
    :returns: The dictionary with the storage configuration for the ``core.psql_dos`` storage plugin.
    """

    def factory(
        database_name: str | None = None, database_username: str | None = None, database_password: str | None = None
    ) -> dict[str, t.Any]:
        storage_config: dict[str, t.Any] = postgres_cluster.create_database(
            database_name=database_name,
            database_username=database_username,
            database_password=database_password,
        )
        storage_config['repository_uri'] = f'file://{tmp_path_factory.mktemp("repository")}'

        return storage_config

    return factory


@pytest.fixture(scope='session')
def config_sqlite_dos(
    tmp_path_factory: pytest.TempPathFactory,
) -> t.Callable[[str | pathlib.Path | None], dict[str, t.Any]]:
    """Return a profile configuration for the :class:`~aiida.storage.sqlite_dos.backend.SqliteDosStorage`.

    The factory has the following signature to allow further configuring the database that is created:

    :param filepath: Optional path to the sqlite database file.
    :returns: The dictionary with the storage configuration for the ``core.sqlite_dos`` storage plugin.
    """

    def factory(filepath: str | pathlib.Path | None = None) -> dict[str, t.Any]:
        return {'filepath': str(filepath or tmp_path_factory.mktemp('test_sqlite_dos_storage'))}

    return factory



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This is the sqlite DB schema, coresponding to the `main_0000` revision of the `sqlite_zip` backend,
see: `versions/main_0000_initial.py`

For normal operation of the archive,
we auto-generate the schema from the models in ``aiida.storage.psql_dos.models``.
However, when migrating an archive from the old format, we require a fixed revision of the schema.

The only difference between the PostGreSQL schema and SQLite one,
is the replacement of ``JSONB`` with ``JSON``, and ``UUID`` with ``CHAR(32)``.
"""

from sqlalchemy import ForeignKey, MetaData, orm
from sqlalchemy.dialects.sqlite import JSON
from sqlalchemy.schema import Column, UniqueConstraint
from sqlalchemy.types import CHAR, Boolean, DateTime, Integer, String, Text

from aiida.common import timezone
from aiida.common.utils import get_new_uuid

# see https://alembic.sqlalchemy.org/en/latest/naming.html
naming_convention = (
    ('pk', '%(table_name)s_pkey'),
    ('ix', 'ix_%(table_name)s_%(column_0_N_label)s'),
    ('uq', 'uq_%(table_name)s_%(column_0_N_name)s'),
    ('ck', 'ck_%(table_name)s_%(constraint_name)s'),
    ('fk', 'fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s'),
)

ArchiveV1Base = orm.declarative_base(metadata=MetaData(naming_convention=dict(naming_convention)))


class DbAuthInfo(ArchiveV1Base):
    """Class that keeps the authentication data."""

    __tablename__ = 'db_dbauthinfo'
    __table_args__ = (UniqueConstraint('aiidauser_id', 'dbcomputer_id'),)

    id = Column(Integer, primary_key=True)
    aiidauser_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True,
    )
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True,
    )
    _metadata = Column('metadata', JSON, default=dict, nullable=True)
    auth_params = Column(JSON, default=dict, nullable=True)
    enabled = Column(Boolean, default=True, nullable=True)


class DbComment(ArchiveV1Base):
    """Class to store comments."""

    __tablename__ = 'db_dbcomment'

    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    dbnode_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True,
    )
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    mtime = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=True,
        index=True,
    )
    content = Column(Text, default='', nullable=True)


class DbComputer(ArchiveV1Base):
    """Class to store computers."""

    __tablename__ = 'db_dbcomputer'

    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    label = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255), default='', nullable=True)
    description = Column(Text, default='', nullable=True)
    scheduler_type = Column(String(255), default='', nullable=True)
    transport_type = Column(String(255), default='', nullable=True)
    _metadata = Column('metadata', JSON, default=dict, nullable=True)


class DbGroupNodes(ArchiveV1Base):
    """Class to store join table for group -> nodes."""

    __tablename__ = 'db_dbgroup_dbnodes'
    __table_args__ = (UniqueConstraint('dbgroup_id', 'dbnode_id'),)

    id = Column(Integer, primary_key=True)
    dbnode_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )
    dbgroup_id = Column(
        Integer, ForeignKey('db_dbgroup.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )


class DbGroup(ArchiveV1Base):
    """Class to store groups."""

    __tablename__ = 'db_dbgroup'
    __table_args__ = (UniqueConstraint('label', 'type_string'),)

    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    label = Column(String(255), nullable=False, index=True)
    type_string = Column(String(255), default='', nullable=True, index=True)
    time = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    description = Column(Text, default='', nullable=True)
    extras = Column(JSON, default=dict, nullable=False)
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )


class DbLog(ArchiveV1Base):
    """Class to store logs."""

    __tablename__ = 'db_dblog'

    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    time = Column(DateTime(timezone=True), default=timezone.now, nullable=True)
    loggername = Column(String(255), default='', nullable=True, index=True)
    levelname = Column(String(50), default='', nullable=True, index=True)
    dbnode_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED', ondelete='CASCADE'),
        nullable=False,
        index=True,
    )
    message = Column(Text(), default='', nullable=True)
    _metadata = Column('metadata', JSON, default=dict, nullable=True)


class DbNode(ArchiveV1Base):
    """Class to store nodes."""

    __tablename__ = 'db_dbnode'

    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), default=get_new_uuid, nullable=False, unique=True)
    node_type = Column(String(255), default='', nullable=False, index=True)
    process_type = Column(String(255), index=True)
    label = Column(String(255), default='', index=True, nullable=True)
    description = Column(Text(), default='', nullable=True)
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=True, index=True)
    mtime = Column(DateTime(timezone=True), default=timezone.now, nullable=True, index=True)
    attributes = Column(JSON)
    extras = Column(JSON)
    repository_metadata = Column(JSON, nullable=False, default=dict, server_default='{}')
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially='DEFERRED', ondelete='RESTRICT'),
        nullable=True,
        index=True,
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially='DEFERRED', ondelete='restrict'),
        nullable=False,
        index=True,
    )


class DbLink(ArchiveV1Base):
    """Class to store links between nodes."""

    __tablename__ = 'db_dblink'

    id = Column(Integer, primary_key=True)
    input_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )
    output_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )
    label = Column(String(255), default='', nullable=False, index=True)
    type = Column(String(255), nullable=False, index=True)


class DbUser(ArchiveV1Base):
    """Class to store users."""

    __tablename__ = 'db_dbuser'

    id = Column(Integer, primary_key=True)
    email = Column(String(254), nullable=False, unique=True)
    first_name = Column(String(254), default='', nullable=True)
    last_name = Column(String(254), default='', nullable=True)
    institution = Column(String(254), default='', nullable=True)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Change UUID type and add uniqueness constraints.

Revision ID: django_0018
Revises: django_0017

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from aiida.storage.psql_dos.migrations.utils import ReflectMigrations
from aiida.storage.psql_dos.migrations.utils.duplicate_uuids import verify_uuid_uniqueness

revision = 'django_0018'
down_revision = 'django_0017'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    reflect = ReflectMigrations(op)

    reflect.drop_indexes('db_dbnode', 'uuid')  # db_dbnode_uuid_62e0bf98_like
    for table, unique in (
        ('db_dbcomment', 'db_dbcomment_uuid_49bac08c_uniq'),
        ('db_dbcomputer', 'db_dbcomputer_uuid_f35defa6_uniq'),
        ('db_dbgroup', 'db_dbgroup_uuid_af896177_uniq'),
        ('db_dbnode', None),
        ('db_dbworkflow', 'db_dbworkflow_uuid_08947ee2_uniq'),
    ):
        op.alter_column(
            table,
            'uuid',
            existing_type=sa.VARCHAR(length=36),
            type_=postgresql.UUID(as_uuid=True),
            nullable=False,
            postgresql_using='uuid::uuid',
        )
        if unique:
            verify_uuid_uniqueness(table, op.get_bind())
            op.create_unique_constraint(unique, table, ['uuid'])

    op.create_unique_constraint('db_dbuser_email_30150b7e_uniq', 'db_dbuser', ['email'])
    op.create_index(
        'db_dbuser_email_30150b7e_like',
        'db_dbuser',
        ['email'],
        postgresql_using='btree',
        postgresql_ops={'email': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0018.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility module with mapper objects that map database entities projections on attributes and labels."""

from . import formatting


class ProjectionMapper:
    """Class to map projection names from the CLI to entity labels, attributes and formatters.

    The command line interface will often have to display database entities and their attributes. The names of
    the attributes exposed on the CLI do not always match one-to-one with the attributes in the ORM and often
    they need to be formatted for the screen in some way. Additionally, for commands that display lists of entries,
    often a header needs to be printed with a label for each attribute, which also are not necessarily identical.

    For any given entity, the CLI typically exposes a set of projections, which are the keywords to reference certain
    attributes. This mapper class serves to map these projections onto the corresponding label and attribute names, as
    well as formatter functions to format the attribute values into strings, suitable to be printed by the CLI.
    """

    _valid_projections = []

    def __init__(self, projection_labels=None, projection_attributes=None, projection_formatters=None):
        """Construct new instance."""
        if not self._valid_projections:
            raise NotImplementedError('no valid projections were specified by the sub class')

        self._projection_labels = {}
        self._projection_attributes = {}
        self._projection_formatters = {}

        if projection_labels is not None:
            for projection in self._valid_projections:
                try:
                    self._projection_labels[projection] = projection_labels[projection]
                except KeyError:
                    self._projection_labels[projection] = projection.replace('_', ' ').capitalize()

        if projection_attributes is not None:
            for projection in self._valid_projections:
                try:
                    self._projection_attributes[projection] = projection_attributes[projection]
                except KeyError:
                    self._projection_attributes[projection] = projection

        if projection_formatters is not None:
            for projection in self._valid_projections:
                try:
                    self._projection_formatters[projection] = projection_formatters[projection]
                except KeyError:
                    attribute = self._projection_attributes[projection]
                    self._projection_formatters[projection] = lambda value, attribute=attribute: value[attribute]

    @property
    def valid_projections(self):
        return self._valid_projections

    def get_label(self, projection):
        return self._projection_labels[projection]

    def get_attribute(self, projection):
        return self._projection_attributes[projection]

    def get_formatter(self, projection):
        return self._projection_formatters[projection]

    def format(self, projection, value):
        return self.get_formatter(projection)(value)


class CalculationProjectionMapper(ProjectionMapper):
    """The CLI projection mapper for Calculation derived entities."""

    def __init__(self, projections, projection_labels=None, projection_attributes=None, projection_formatters=None):
        from aiida.orm import ProcessNode
        from aiida.orm.nodes.caching import NodeCaching
        from aiida.orm.utils.mixins import Sealable

        self._valid_projections = projections

        sealed_key = f'attributes.{Sealable.SEALED_KEY}'
        job_state_key = 'attributes.state'
        scheduler_state_key = 'attributes.scheduler_state'
        process_paused_key = f'attributes.{ProcessNode.PROCESS_PAUSED_KEY}'
        process_label_key = f'attributes.{ProcessNode.PROCESS_LABEL_KEY}'
        process_state_key = f'attributes.{ProcessNode.PROCESS_STATE_KEY}'
        process_status_key = f'attributes.{ProcessNode.PROCESS_STATUS_KEY}'
        exit_status_key = f'attributes.{ProcessNode.EXIT_STATUS_KEY}'
        exit_message_key = f'attributes.{ProcessNode.EXIT_MESSAGE_KEY}'
        exception_key = f'attributes.{ProcessNode.EXCEPTION_KEY}'
        cached_from_key = f'extras.{NodeCaching.CACHED_FROM_KEY}'

        default_labels = {
            'pk': 'PK',
            'uuid': 'UUID',
            'ctime': 'Created',
            'mtime': 'Modified',
            'state': 'Process State',
            'cached': '\u267b',
        }

        default_attributes = {
            'pk': 'id',
            'job_state': job_state_key,
            'scheduler_state': scheduler_state_key,
            'sealed': sealed_key,
            'paused': process_paused_key,
            'process_label': process_label_key,
            'process_state': process_state_key,
            'process_status': process_status_key,
            'exit_status': exit_status_key,
            'exit_message': exit_message_key,
            'exception': exception_key,
            'cached': cached_from_key,
            'cached_from': cached_from_key,
        }

        default_formatters = {
            'ctime': lambda value: formatting.format_relative_time(value['ctime']),
            'mtime': lambda value: formatting.format_relative_time(value['mtime']),
            'state': lambda v: formatting.format_state(v[process_state_key], v[process_paused_key], v[exit_status_key]),
            'process_state': lambda value: formatting.format_process_state(value[process_state_key]),
            'sealed': lambda value: formatting.format_sealed(value[sealed_key]),
            'cached': lambda value: '\u2714' if value[cached_from_key] else '',
        }

        if projection_labels is not None:
            for projection, label in projection_labels.items():
                if projection not in self.valid_projections:
                    raise ValueError(f'{projection} is not a valid projection')
                else:
                    default_labels[projection] = label

        if projection_attributes is not None:
            for projection, attribute in projection_attributes.items():
                if projection not in self.valid_projections:
                    raise ValueError(f'{projection} is not a valid projection')
                else:
                    default_attributes[projection] = attribute

        if projection_formatters is not None:
            for projection, formatter in projection_formatters.items():
                if projection not in self.valid_projections:
                    raise ValueError(f'{projection} is not a valid projection')
                else:
                    default_formatters[projection] = formatter

        super().__init__(default_labels, default_attributes, default_formatters)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Shared resources for the archive."""

import urllib.parse
import urllib.request
from html.parser import HTMLParser
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type

from aiida.orm import AuthInfo, Comment, Computer, Entity, Group, Log, Node, User
from aiida.orm.entities import EntityTypes

# Mapping from entity names to AiiDA classes
entity_type_to_orm: Dict[EntityTypes, Type[Entity]] = {
    EntityTypes.AUTHINFO: AuthInfo,
    EntityTypes.GROUP: Group,
    EntityTypes.COMPUTER: Computer,
    EntityTypes.USER: User,
    EntityTypes.LOG: Log,
    EntityTypes.NODE: Node,
    EntityTypes.COMMENT: Comment,
}


def batch_iter(
    iterable: Iterable[Any], size: int, transform: Optional[Callable[[Any], Any]] = None
) -> Iterable[Tuple[int, List[Any]]]:
    """Yield an iterable in batches of a set number of items.

    Note, the final yield may be less than this size.

    :param transform: a transform to apply to each item
    :returns: (number of items, list of items)
    """
    transform = transform or (lambda x: x)
    current = []
    length = 0
    for item in iterable:
        current.append(transform(item))
        length += 1
        if length >= size:
            yield length, current
            current = []
            length = 0
    if current:
        yield length, current


class HTMLGetLinksParser(HTMLParser):
    """If a filter_extension is passed, only links with extension matching
    the given one will be returned.
    """

    def __init__(self, filter_extension=None):
        self.filter_extension = filter_extension
        self.links = []
        super().__init__()

    def handle_starttag(self, tag, attrs):
        """Store the urls encountered, if they match the request."""
        if tag == 'a':
            for key, value in attrs:
                if key == 'href':
                    if self.filter_extension is None or value.endswith(f'.{self.filter_extension}'):
                        self.links.append(value)

    def get_links(self):
        """Return the links that were found during the parsing phase."""
        return self.links


def get_valid_import_links(url):
    """Open the given URL, parse the HTML and return a list of valid links where
    the link file has a .aiida extension.
    """
    with urllib.request.urlopen(url) as request:
        parser = HTMLGetLinksParser(filter_extension='aiida')
        parser.feed(request.read().decode('utf8'))

    return_urls = []

    for link in parser.get_links():
        return_urls.append(urllib.parse.urljoin(request.geturl(), link))

    return return_urls



"""Implementation of the ``AbstractRepositoryBackend`` using the ``disk-objectstore`` as the backend."""

import contextlib
import dataclasses
import shutil
import typing as t

from aiida.common.lang import type_check
from aiida.storage.log import STORAGE_LOGGER

from .abstract import AbstractRepositoryBackend

if t.TYPE_CHECKING:
    from disk_objectstore import Container

__all__ = ('DiskObjectStoreRepositoryBackend',)

BYTES_TO_MB = 1 / 1024**2

logger = STORAGE_LOGGER.getChild('disk_object_store')


class DiskObjectStoreRepositoryBackend(AbstractRepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` using the ``disk-object-store`` as the backend.

    .. note:: For certain methods, the container may create a sessions which should be closed after the operation is
        done to make sure the connection to the underlying sqlite database is closed. The best way is to accomplish this
        is by using the container as a context manager, which will automatically call the ``close`` method when it exits
        which ensures the session being closed. Note that not all methods may open the session and so need closing it,
        but to be on the safe side, we put every use of the container in a context manager. If no session is created,
        the ``close`` method is essentially a no-op.

    """

    def __init__(self, container: 'Container'):
        if not t.TYPE_CHECKING:
            from disk_objectstore import Container
        type_check(container, Container)
        self._container = container

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        if self.is_initialised:
            with self._container as container:
                return f'DiskObjectStoreRepository: {container.container_id} | {container.get_folder()}'
        return 'DiskObjectStoreRepository: <uninitialised>'

    @property
    def uuid(self) -> t.Optional[str]:
        """Return the unique identifier of the repository."""
        if not self.is_initialised:
            return None
        with self._container as container:
            return container.container_id

    @property
    def key_format(self) -> t.Optional[str]:
        with self._container as container:
            return container.hash_type

    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """
        with self._container as container:
            container.init_container(**kwargs)

    @property
    def is_initialised(self) -> bool:
        """Return whether the repository has been initialised."""
        with self._container as container:
            return container.is_initialised

    def erase(self):
        """Delete the repository itself and all its contents."""
        try:
            with self._container as container:
                shutil.rmtree(container.get_folder())
        except FileNotFoundError:
            pass

    def _put_object_from_filelike(self, handle: t.BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        with self._container as container:
            return container.add_streamed_object(handle)

    def has_objects(self, keys: t.List[str]) -> t.List[bool]:
        with self._container as container:
            return container.has_objects(keys)

    @contextlib.contextmanager
    def open(self, key: str) -> t.Iterator[t.BinaryIO]:
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param key: fully qualified identifier for the object within the repository.
        :return: yield a byte stream object.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        super().open(key)

        with self._container as container:
            with container.get_object_stream(key) as handle:
                yield handle  # type: ignore[misc]

    def iter_object_streams(self, keys: t.List[str]) -> t.Iterator[t.Tuple[str, t.BinaryIO]]:
        with self._container.get_objects_stream_and_meta(keys) as triplets:
            for key, stream, _ in triplets:
                assert stream is not None
                yield key, stream  # type: ignore[misc]

    def delete_objects(self, keys: t.List[str]) -> None:
        super().delete_objects(keys)
        with self._container as container:
            container.delete_objects(keys)

    def list_objects(self) -> t.Iterable[str]:
        with self._container as container:
            return container.list_all_objects()

    def get_object_hash(self, key: str) -> str:
        """Return the SHA-256 hash of an object stored under the given key.

        .. important::
            A SHA-256 hash should always be returned,
            to ensure consistency across different repository implementations.

        :param key: fully qualified identifier for the object within the repository.
        :raise FileNotFoundError: if the file does not exist.

        """
        if not self.has_object(key):
            raise FileNotFoundError(key)
        with self._container as container:
            if container.hash_type != 'sha256':
                return super().get_object_hash(key)
        return key

    def maintain(  # type: ignore[override]
        self,
        dry_run: bool = False,
        live: bool = True,
        pack_loose: t.Optional[bool] = None,
        do_repack: t.Optional[bool] = None,
        clean_storage: t.Optional[bool] = None,
        do_vacuum: t.Optional[bool] = None,
        compress: bool = False,
    ) -> dict:
        """Performs maintenance operations.

        :param live: if True, will only perform operations that are safe to do while the repository is in use.
        :param pack_loose: flag for forcing the packing of loose files.
        :param do_repack: flag for forcing the re-packing of already packed files.
        :param clean_storage: flag for forcing the cleaning of soft-deleted files from the repository.
        :param do_vacuum: flag for forcing the vacuuming of the internal database when cleaning the repository.
        :param compress: flag for compressing the data when packing loose files. Set to ``Compress.AUTO`` if ``True``.
        :return: a dictionary with information on the operations performed.
        """
        from disk_objectstore import CompressMode

        if live and (do_repack or clean_storage or do_vacuum):
            overrides = {'do_repack': do_repack, 'clean_storage': clean_storage, 'do_vacuum': do_vacuum}
            keys = ', '.join([key for key, override in overrides.items() if override is True])  # type: ignore
            raise ValueError(f'The following overrides were enabled but cannot be if `live=True`: {keys}')

        pack_loose = True if pack_loose is None else pack_loose

        if compress is True:
            compress = CompressMode.AUTO

        if live:
            do_repack = False
            clean_storage = True if clean_storage is None else clean_storage
            do_vacuum = False
        else:
            do_repack = True if do_repack is None else do_repack
            clean_storage = True if clean_storage is None else clean_storage
            do_vacuum = True if do_vacuum is None else do_vacuum

        with self._container as container:
            if pack_loose:
                files_numb = container.count_objects().loose
                files_size = container.get_total_size().total_size_loose * BYTES_TO_MB
                logger.report(f'Packing all loose files ({files_numb} files occupying {files_size} MB) ...')
                if not dry_run:
                    container.pack_all_loose(compress=compress)

            if do_repack:
                files_numb = container.count_objects().packed
                files_size = container.get_total_size().total_size_packfiles_on_disk * BYTES_TO_MB
                logger.report(f'Re-packing all pack files ({files_numb} files in packs, occupying {files_size} MB) ...')
                if not dry_run:
                    container.repack()

            if clean_storage:
                logger.report(f'Cleaning the repository database (with `vacuum={do_vacuum}`) ...')
                if not dry_run:
                    container.clean_storage(vacuum=do_vacuum)

    def get_info(  # type: ignore[override]
        self,
        detailed=False,
    ) -> t.Dict[str, t.Union[int, str, t.Dict[str, int], t.Dict[str, float]]]:
        """Return information on configuration and content of the repository."""
        output_info: t.Dict[str, t.Union[int, str, t.Dict[str, int], t.Dict[str, float]]] = {}

        with self._container as container:
            output_info['SHA-hash algorithm'] = container.hash_type
            output_info['Compression algorithm'] = container.compression_algorithm
            output_info['Objects'] = dataclasses.asdict(container.count_objects())

            if not detailed:
                return output_info

            output_info['Size (MB)'] = {
                k: float(f'{v * BYTES_TO_MB:.2f}') for k, v in dataclasses.asdict(container.get_total_size()).items()
            }

        return output_info



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a folder on a file system."""

from __future__ import annotations

import contextlib
import io
import pathlib
import typing as t

from .data import Data

if t.TYPE_CHECKING:
    from aiida.repository import File


__all__ = ('FolderData',)

FilePath = t.Union[str, pathlib.PurePosixPath]


class FolderData(Data):
    """`Data` sub class to represent a folder on a file system."""

    def __init__(self, **kwargs):
        """Construct a new `FolderData` to which any files and folders can be added.

        Use the `tree` keyword to simply wrap a directory:

            folder = FolderData(tree='/absolute/path/to/directory')

        Alternatively, one can construct the node first and then use the various repository methods to add objects:

            folder = FolderData()
            folder.put_object_from_tree('/absolute/path/to/directory')
            folder.put_object_from_filepath('/absolute/path/to/file.txt')
            folder.put_object_from_filelike(filelike_object)

        :param tree: absolute path to a folder to wrap
        :type tree: str
        """
        tree = kwargs.pop('tree', None)
        super().__init__(**kwargs)
        if tree:
            self.base.repository.put_object_from_tree(tree)

    def list_objects(self, path: str | None = None) -> list[File]:
        """Return a list of the objects contained in this repository sorted by name, optionally in given sub directory.

        :param path: optional relative path inside the repository whose objects to list.
        :return: a list of `File` named tuples representing the objects present in directory with the given key.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        return self.base.repository.list_objects(path)

    def list_object_names(self, path: str | None = None) -> list[str]:
        """Return a sorted list of the object names contained in this repository, optionally in the given sub directory.

        :param path: optional relative path inside the repository whose objects to list.
        :return: a list of `File` named tuples representing the objects present in directory with the given key.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        return self.base.repository.list_object_names(path)

    @t.overload
    @contextlib.contextmanager
    def open(self, path: FilePath, mode: t.Literal['r']) -> t.Iterator[t.TextIO]: ...

    @t.overload
    @contextlib.contextmanager
    def open(self, path: FilePath, mode: t.Literal['rb']) -> t.Iterator[t.BinaryIO]: ...

    @contextlib.contextmanager
    def open(self, path: FilePath, mode: t.Literal['r', 'rb'] = 'r') -> t.Iterator[t.BinaryIO] | t.Iterator[t.TextIO]:
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param path: the relative path of the object within the repository.
        :return: yield a byte stream object.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        with self.base.repository.open(path, mode) as handle:
            yield handle

    @contextlib.contextmanager
    def as_path(self, path: FilePath | None = None) -> t.Iterator[pathlib.Path]:
        """Make the contents of the repository available as a normal filepath on the local file system.

        :param path: optional relative path of the object within the repository.
        :return: the filepath of the content of the repository or object if ``path`` is specified.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        """
        with self.base.repository.as_path(path) as filepath:
            yield filepath

    def get_object(self, path: FilePath | None = None) -> File:
        """Return the object at the given path.

        :param path: the relative path of the object within the repository.
        :return: the `File` representing the object located at the given relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        """
        return self.base.repository.get_object(path)

    @t.overload
    def get_object_content(self, path: str, mode: t.Literal['r']) -> str: ...

    @t.overload
    def get_object_content(self, path: str, mode: t.Literal['rb']) -> bytes: ...

    def get_object_content(self, path: str, mode: t.Literal['r', 'rb'] = 'r') -> str | bytes:
        """Return the content of a object identified by key.

        :param path: the relative path of the object within the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        return self.base.repository.get_object_content(path, mode)

    def put_object_from_bytes(self, content: bytes, path: str) -> None:
        """Store the given content in the repository at the given path.

        :param path: the relative path where to store the object in the repository.
        :param content: the content to store.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileExistsError: if an object already exists at the given path.
        """
        return self.base.repository.put_object_from_bytes(content, path)

    def put_object_from_filelike(self, handle: io.BufferedReader, path: str) -> None:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        return self.base.repository.put_object_from_filelike(handle, path)

    def put_object_from_file(self, filepath: str, path: str) -> None:
        """Store a new object under `path` with contents of the file located at `filepath` on the local file system.

        :param filepath: absolute path of file whose contents to copy to the repository
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path, or the handle is not a byte stream.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        return self.base.repository.put_object_from_file(filepath, path)

    def put_object_from_tree(self, filepath: str, path: str | None = None) -> None:
        """Store the entire contents of `filepath` on the local file system in the repository with under given `path`.

        :param filepath: absolute path of the directory whose contents to copy to the repository.
        :param path: the relative path where to store the objects in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        return self.base.repository.put_object_from_tree(filepath, path)

    def walk(self, path: FilePath | None = None) -> t.Iterable[tuple[pathlib.PurePosixPath, list[str], list[str]]]:
        """Walk over the directories and files contained within this repository.

        .. note:: the order of the dirname and filename lists that are returned is not necessarily sorted. This is in
            line with the ``os.walk`` implementation where the order depends on the underlying file system used.

        :param path: the relative path of the directory within the repository whose contents to walk.
        :return: tuples of root, dirnames and filenames just like ``os.walk``, with the exception that the root path is
            always relative with respect to the repository root, instead of an absolute path and it is an instance of
            ``pathlib.PurePosixPath`` instead of a normal string
        """
        yield from self.base.repository.walk(path)

    def glob(self) -> t.Iterable[pathlib.PurePosixPath]:
        """Yield a recursive list of all paths (files and directories)."""
        yield from self.base.repository.glob()

    def copy_tree(self, target: str | pathlib.Path, path: FilePath | None = None) -> None:
        """Copy the contents of the entire node repository to another location on the local file system.

        :param target: absolute path of the directory where to copy the contents to.
        :param path: optional relative path whose contents to copy.
        """
        self.base.repository.copy_tree(target, path)

    def delete_object(self, path: str) -> None:
        """Delete the object from the repository.

        :param key: fully qualified identifier for the object within the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be deleted.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.base.repository.delete_object(path)

    def erase(self) -> None:
        """Delete all objects from the repository.

        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.base.repository.erase()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the `AuthInfo` ORM class."""

from typing import TYPE_CHECKING, Any, Dict, Optional, Type

from aiida.common import exceptions
from aiida.manage import get_manager
from aiida.plugins import TransportFactory

from . import entities, users
from .fields import add_field

if TYPE_CHECKING:
    from aiida.orm import Computer, User
    from aiida.orm.implementation import StorageBackend
    from aiida.orm.implementation.authinfos import BackendAuthInfo  # noqa: F401
    from aiida.transports import Transport

__all__ = ('AuthInfo',)


class AuthInfoCollection(entities.Collection['AuthInfo']):
    """The collection of `AuthInfo` entries."""

    @staticmethod
    def _entity_base_cls() -> Type['AuthInfo']:
        return AuthInfo

    def delete(self, pk: int) -> None:
        """Delete an entry from the collection.

        :param pk: the pk of the entry to delete
        """
        self._backend.authinfos.delete(pk)


class AuthInfo(entities.Entity['BackendAuthInfo', AuthInfoCollection]):
    """ORM class that models the authorization information that allows a `User` to connect to a `Computer`."""

    _CLS_COLLECTION = AuthInfoCollection

    __qb_fields__ = [
        add_field(
            'enabled',
            dtype=bool,
            is_attribute=False,
            doc='Whether the instance is enabled',
        ),
        add_field(
            'auth_params',
            dtype=Dict[str, Any],
            is_attribute=False,
            doc='Dictionary of authentication parameters',
        ),
        add_field(
            'metadata',
            dtype=Dict[str, Any],
            is_attribute=False,
            doc='Dictionary of metadata',
        ),
        add_field(
            'computer_pk',
            dtype=int,
            is_attribute=False,
            doc='The PK of the computer',
        ),
        add_field(
            'user_pk',
            dtype=int,
            is_attribute=False,
            doc='The PK of the user',
        ),
    ]

    PROPERTY_WORKDIR = 'workdir'

    def __init__(self, computer: 'Computer', user: 'User', backend: Optional['StorageBackend'] = None) -> None:
        """Create an `AuthInfo` instance for the given computer and user.

        :param computer: a `Computer` instance
        :param user: a `User` instance
        :param backend: the backend to use for the instance, or use the default backend if None
        """
        backend = backend or get_manager().get_profile_storage()
        model = backend.authinfos.create(computer=computer.backend_entity, user=user.backend_entity)
        super().__init__(model)

    def __str__(self) -> str:
        if self.enabled:
            return f'AuthInfo for {self.user.email} on {self.computer.label}'

        return f'AuthInfo for {self.user.email} on {self.computer.label} [DISABLED]'

    @property
    def enabled(self) -> bool:
        """Return whether this instance is enabled.

        :return: True if enabled, False otherwise
        """
        return self._backend_entity.enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        """Set the enabled state

        :param enabled: boolean, True to enable the instance, False to disable it
        """
        self._backend_entity.enabled = enabled

    @property
    def computer(self) -> 'Computer':
        """Return the computer associated with this instance."""
        from . import computers

        return entities.from_backend_entity(computers.Computer, self._backend_entity.computer)

    @property
    def user(self) -> 'User':
        """Return the user associated with this instance."""
        return entities.from_backend_entity(users.User, self._backend_entity.user)

    def get_auth_params(self) -> Dict[str, Any]:
        """Return the dictionary of authentication parameters

        :return: a dictionary with authentication parameters
        """
        return self._backend_entity.get_auth_params()

    def set_auth_params(self, auth_params: Dict[str, Any]) -> None:
        """Set the dictionary of authentication parameters

        :param auth_params: a dictionary with authentication parameters
        """
        self._backend_entity.set_auth_params(auth_params)

    def get_metadata(self) -> Dict[str, Any]:
        """Return the dictionary of metadata

        :return: a dictionary with metadata
        """
        return self._backend_entity.get_metadata()

    def set_metadata(self, metadata: Dict[str, Any]) -> None:
        """Set the dictionary of metadata

        :param metadata: a dictionary with metadata
        """
        self._backend_entity.set_metadata(metadata)

    def get_workdir(self) -> str:
        """Return the working directory.

        If no explicit work directory is set for this instance, the working directory of the computer will be returned.

        :return: the working directory
        """
        try:
            return self.get_metadata()[self.PROPERTY_WORKDIR]
        except KeyError:
            return self.computer.get_workdir()

    def get_transport(self) -> 'Transport':
        """Return a fully configured transport that can be used to connect to the computer set for this instance."""
        computer = self.computer
        transport_type = computer.transport_type

        try:
            transport_class = TransportFactory(transport_type)
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError(f'transport type `{transport_type}` could not be loaded: {exception}')

        return transport_class(machine=computer.hostname, **self.get_auth_params())



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Create an old style node attribute/extra, via the `db_dbattribute`/`db_dbextra` tables.

Adapted from: `aiida/backends/djsite/db/migrations/__init__.py`
"""

from __future__ import annotations

import datetime
import json

from aiida.common.exceptions import ValidationError
from aiida.common.timezone import make_aware


def create_rows(key: str, value, node_id: int) -> list[dict]:
    """Create an old style node attribute/extra, via the `db_dbattribute`/`db_dbextra` tables.

    :note: No hits are done on the DB, in particular no check is done
        on the existence of the given nodes.

    :param key: a string with the key to create (can contain the
        separator self._sep if this is a sub-attribute: indeed, this
        function calls itself recursively)
    :param value: the value to store (a basic data type or a list or a dict)
    :param node_id: the node id to store the attribute/extra

    :return: A list of column name -> value dictionaries, with which to instantiate database rows
    """
    list_to_return = []

    columns = {
        'key': key,
        'dbnode_id': node_id,
        'datatype': 'none',
        'tval': '',
        'bval': None,
        'ival': None,
        'fval': None,
        'dval': None,
    }

    if isinstance(value, bool):
        columns['datatype'] = 'bool'
        columns['bval'] = value

    elif isinstance(value, int):
        columns['datatype'] = 'int'
        columns['ival'] = value

    elif isinstance(value, float):
        columns['datatype'] = 'float'
        columns['fval'] = value
        columns['tval'] = ''

    elif isinstance(value, str):
        columns['datatype'] = 'txt'
        columns['tval'] = value

    elif isinstance(value, datetime.datetime):
        columns['datatype'] = 'date'
        columns['dval'] = make_aware(value)

    elif isinstance(value, (list, tuple)):
        columns['datatype'] = 'list'
        columns['ival'] = len(value)

        for i, subv in enumerate(value):
            # I do not need get_or_create here, because
            # above I deleted all children (and I
            # expect no concurrency)
            # NOTE: I do not pass other_attribs
            list_to_return.extend(create_rows(f'{key}.{i:d}', subv, node_id))

    elif isinstance(value, dict):
        columns['datatype'] = 'dict'
        columns['ival'] = len(value)

        for subk, subv in value.items():
            if not isinstance(key, str) or not key:  # type: ignore[redundant-expr]
                raise ValidationError('The key must be a non-empty string.')
            if '.' in key:
                raise ValidationError(
                    "The separator symbol '.' cannot be present in the key of attributes, extras, etc."
                )
            list_to_return.extend(create_rows(f'{key}.{subk}', subv, node_id))
    else:
        try:
            jsondata = json.dumps(value)
        except TypeError:
            raise ValueError(
                f'Unable to store the value: it must be either a basic datatype, or json-serializable: {value}'
            ) from TypeError

        columns['datatype'] = 'json'
        columns['tval'] = jsondata

    # create attr row and add to list_to_return
    list_to_return.append(columns)

    return list_to_return



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Comment objects and functions"""

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Type

from aiida.manage import get_manager

from . import entities, users
from .fields import add_field

if TYPE_CHECKING:
    from aiida.orm import Node, User
    from aiida.orm.implementation import StorageBackend

__all__ = ('Comment',)


class CommentCollection(entities.Collection['Comment']):
    """The collection of Comment entries."""

    @staticmethod
    def _entity_base_cls() -> Type['Comment']:
        return Comment

    def delete(self, pk: int) -> None:
        """Remove a Comment from the collection with the given id

        :param pk: the id of the comment to delete

        :raises TypeError: if ``comment_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Comment with ID ``comment_id`` is not found
        """
        self._backend.comments.delete(pk)

    def delete_all(self) -> None:
        """Delete all Comments from the Collection

        :raises `~aiida.common.exceptions.IntegrityError`: if all Comments could not be deleted
        """
        self._backend.comments.delete_all()

    def delete_many(self, filters: dict) -> List[int]:
        """Delete Comments from the Collection based on ``filters``

        :param filters: similar to QueryBuilder filter

        :return: (former) ``PK`` s of deleted Comments

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """
        return self._backend.comments.delete_many(filters)


class Comment(entities.Entity['BackendComment', CommentCollection]):
    """Base class to map a DbComment that represents a comment attached to a certain Node."""

    _CLS_COLLECTION = CommentCollection

    __qb_fields__ = [
        add_field(
            'uuid',
            dtype=str,
            is_attribute=False,
            doc='The UUID of the comment',
        ),
        add_field(
            'ctime',
            dtype=datetime,
            is_attribute=False,
            doc='Creation time of the comment',
        ),
        add_field(
            'mtime',
            dtype=datetime,
            is_attribute=False,
            doc='Modified time of the comment',
        ),
        add_field(
            'content',
            dtype=str,
            is_attribute=False,
            doc='Content of the comment',
        ),
        add_field(
            'user_pk',
            dtype=int,
            is_attribute=False,
            doc='User PK that created the comment',
        ),
        add_field(
            'node_pk',
            dtype=int,
            is_attribute=False,
            doc='Node PK that the comment is attached to',
        ),
    ]

    def __init__(
        self, node: 'Node', user: 'User', content: Optional[str] = None, backend: Optional['StorageBackend'] = None
    ):
        """Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :param backend: the backend to use for the instance, or use the default backend if None

        :return: a Comment object associated to the given node and user
        """
        backend = backend or get_manager().get_profile_storage()
        model = backend.comments.create(node=node.backend_entity, user=user.backend_entity, content=content)
        super().__init__(model)

    def __str__(self) -> str:
        arguments = [self.uuid, self.node.pk, self.user.email, self.content]
        return 'Comment<{}> for node<{}> and user<{}>: {}'.format(*arguments)

    @property
    def uuid(self) -> str:
        """Return the UUID for this comment.

        This identifier is unique across all entities types and backend instances.

        :return: the entity uuid
        """
        return self._backend_entity.uuid

    @property
    def ctime(self) -> datetime:
        return self._backend_entity.ctime

    @property
    def mtime(self) -> datetime:
        return self._backend_entity.mtime

    def set_mtime(self, value: datetime) -> None:
        return self._backend_entity.set_mtime(value)

    @property
    def node(self) -> 'Node':
        return self._backend_entity.node

    @property
    def user(self) -> 'User':
        return entities.from_backend_entity(users.User, self._backend_entity.user)

    def set_user(self, value: 'User') -> None:
        self._backend_entity.user = value.backend_entity

    @property
    def content(self) -> str:
        return self._backend_entity.content

    def set_content(self, value: str) -> None:
        return self._backend_entity.set_content(value)
